
__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import time
import queue
import collections

import threading     as mt

import radical.utils as ru

from . import utils     as rpu
from . import states    as rps
from . import constants as rpc

from .task_description import RAPTOR_MASTER, RAPTOR_WORKER
from .raptor_tasks     import RaptorMaster, RaptorWorker


# bulk callbacks are implemented, but are currently not used nor exposed.
_USE_BULK_CB = False
if os.environ.get('RADICAL_PILOT_BULK_CB', '').lower() in ['true', 'yes', '1']:
    _USE_BULK_CB = True


# ------------------------------------------------------------------------------
#
class TaskManager(rpu.ClientComponent):
    """
    A TaskManager manages :class:`radical.pilot.Task` instances which
    represent the **executable** workload in RADICAL-Pilot. A TaskManager
    connects the Tasks with one or more :class:`Pilot` instances (which
    represent the workload **executors** in RADICAL-Pilot) and a **scheduler**
    which determines which :class:`Task` gets executed on which
    :class:`Pilot`.

    Example::

        s = rp.Session()

        pm = rp.PilotManager(session=s)

        pd = rp.PilotDescription()
        pd.resource = "futuregrid.alamo"
        pd.cores = 16

        p1 = pm.submit_pilots(pd) # create first pilot with 16 cores
        p2 = pm.submit_pilots(pd) # create second pilot with 16 cores

        # Create a workload of 128 '/bin/sleep' tasks
        tasks = []
        for task_count in range(0, 128):
            t = rp.TaskDescription()
            t.executable = "/bin/sleep"
            t.arguments = ['60']
            tasks.append(t)

        # Combine the two pilots, the workload and a scheduler via
        # a TaskManager.
        tm = rp.TaskManager(session=session, scheduler=rp.SCHEDULER_ROUND_ROBIN)
        tm.add_pilot(p1)
        tm.submit_tasks(tasks)


    The task manager can issue notification on task state changes.  Whenever
    state notification arrives, any callback registered for that notification is
    fired.

    Note:
        State notifications can arrive out of order wrt the task state model!

    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, session, cfg='default', scheduler=None):
        """Create a new TaskManager and attaches it to the session.

        Arguments:
            session (radical.pilot.Session): The session instance to use.
            cfg (dict | str): The configuration or name of configuration to use.
            scheduler (str): The name of the scheduler plug-in to use.
            uid (str): ID for unit manager, to be used for reconnect

        Returns:
            radical.pilot.TaskManager: A new `TaskManager` object.

        """

        assert session._role == session._PRIMARY, 'tmgr needs primary session'

        # initialize the base class (with no intent to fork)
        self._uid = ru.generate_id('tmgr.%(item_counter)04d',
                                    ru.ID_CUSTOM, ns=session.uid)

        if not scheduler:
            scheduler = rpc.SCHEDULER_ROUND_ROBIN

        self._pilots      = dict()
        self._pilots_lock = mt.RLock()
        self._tasks       = dict()
        self._tasks_lock  = mt.RLock()
        self._callbacks   = dict()
        self._tcb_lock    = mt.RLock()
        self._terminate   = mt.Event()
        self._closed      = False
        self._task_info   = collections.defaultdict(dict)

        for m in rpc.TMGR_METRICS:
            self._callbacks[m] = dict()

        # NOTE: `name` and `cfg` are overloaded, the user cannot point to
        #       a predefined config and name it at the same time.  This might
        #       be ok for the session, but introduces a minor API inconsistency.
        #
        name = None
        if isinstance(cfg, str):
            name = cfg
            cfg  = None

        cfg                = ru.Config('radical.pilot.tmgr', name=name, cfg=cfg)
        cfg.uid            = self._uid
        cfg.owner          = self._uid
        cfg.sid            = session.uid
        cfg.path           = session.path
        cfg.reg_addr       = session.reg_addr
        cfg.heartbeat      = session.cfg.heartbeat
        cfg.client_sandbox = session._get_client_sandbox()

        super().__init__(cfg, session=session)
        self.start()

        self._log.info('started tmgr %s', self._uid)

        self._rep = self._session._get_reporter(name=self._uid)
        self._rep.info('<<create task manager')

        # overwrite the scheduler from the config file
        if 'tmgr_scheduling' in self._cfg.components:
            self._cfg.components.tmgr_scheduling.scheduler = scheduler

        # create pmgr bridges and components, use session cmgr for that
        self._cmgr = rpu.ComponentManager(cfg.sid, cfg.reg_addr, self._uid)
        self._cmgr.start_bridges(self._cfg.bridges)
        self._cmgr.start_components(self._cfg.components)

        # let session know we exist
        self._session._register_tmgr(self)

        # The output queue is used to forward submitted tasks to the
        # scheduling component.
        self.register_output(rps.TMGR_SCHEDULING_PENDING,
                             rpc.TMGR_SCHEDULING_QUEUE)

        # the tmgr will also collect tasks from the agent again, for output
        # staging and finalization
        if self._cfg.bridges.tmgr_staging_output_queue:
            self._has_sout = True
            self.register_output(rps.TMGR_STAGING_OUTPUT_PENDING,
                                 rpc.TMGR_STAGING_OUTPUT_QUEUE)
        else:
            self._has_sout = False

        # also listen to the state pubsub for task state changes
        self.register_subscriber(rpc.STATE_PUBSUB, self._state_sub_cb)

        # hook into the control pubsub for rpc handling
        self._rpc_queue = queue.Queue()
        ctrl_addr_pub   = self._session._reg['bridges.control_pubsub.addr_pub']
        ctrl_addr_sub   = self._session._reg['bridges.control_pubsub.addr_sub']

        self._ctrl_pub  = ru.zmq.Publisher(rpc.CONTROL_PUBSUB, url=ctrl_addr_pub,
                                           log=self._log, prof=self._prof)

        self._ctrl_sub = ru.zmq.Subscriber(rpc.CONTROL_PUBSUB, url=ctrl_addr_sub,
                                           log=self._log, prof=self._prof,
                                           cb=self._control_cb,
                                           topic=rpc.CONTROL_PUBSUB)

        self._prof.prof('setup_done', uid=self._uid)
        self._rep.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # the manager must not carry bridge and component handles across forks
        ru.atfork(self._atfork_prepare, self._atfork_parent, self._atfork_child)


    # --------------------------------------------------------------------------
    #
    # EnTK forks, make sure we don't carry traces of children across the fork
    #
    def _atfork_prepare(self): pass
    def _atfork_parent(self) : pass
    def _atfork_child(self)  :
        self._bridges    = dict()
        self._components = dict()


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self._cmgr.close()


    # --------------------------------------------------------------------------
    #
    def close(self):
        """Shut down the TaskManager and all its components."""

        # we do not cancel tasks at this point, in case any component or pilot
        # wants to continue to progress task states, which should indeed be
        # independent from the tmgr life cycle.

        if self._closed:
            return

        self._terminate.set()
        self._rep.info('<<close task manager')

        # disable callbacks during shutdown
        with self._tcb_lock:
            self._callbacks = dict()
            for m in rpc.TMGR_METRICS:
                self._callbacks[m] = dict()

        self._cmgr.close()

        self._log.info("Closed TaskManager %s." % self._uid)

        self._closed = True
        self._rep.ok('>>ok\n')

        self.dump()

        self._ctrl_sub.stop()

        super().close()


    # --------------------------------------------------------------------------
    #
    def dump(self, name=None):

        # dump json
        json = self.as_dict()
      # json['_id']   = self.uid
        json['type']  = 'tmgr'
        json['uid']   = self.uid
        json['tasks'] = self._task_info

        if name:
            tgt = '%s/%s.%s.json' % (self._session.path, self.uid, name)
        else:
            tgt = '%s/%s.json' % (self._session.path, self.uid)

        ru.write_json(json, tgt)

    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a dictionary representation of the TaskManager object."""

        ret = {
            'uid': self.uid,
            'cfg': self.cfg
        }

        return ret


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns a string representation of the TaskManager object."""

        return str(self.as_dict())


    # --------------------------------------------------------------------------
    #
    def _pilot_state_cb(self, pilots, state=None):

        if self._terminate.is_set():
            return False

        # we register this callback for pilots added to this tmgr.  It will
        # specifically look out for pilots which complete, and will make sure
        # that all tasks are pulled back into tmgr control if that happens
        # prematurely.
        #
        # If we find tasks which have not completed the agent part of the task
        # state model, we declare them FAILED.  If they can be restarted, we
        # resubmit an identical task, which then will get a new task ID.  This
        # avoids state model confusion (the state model is right now expected to
        # be linear), but is not intuitive for the application (FIXME).
        #
        # FIXME: there is a race with the tmgr scheduler which may, just now,
        #        and before being notified about the pilot's demise, send new
        #        tasks to the pilot.

        # we only look into pilot states when the tmgr is still active
        # FIXME: note that there is a race in that the tmgr can be closed while
        #        we are in the cb.
        # FIXME: `self._closed` is not an `mt.Event`!
        if self._closed:
            self._log.debug('tmgr closed, ignore pilot cb %s',
                            ['%s:%s' % (p.uid, p.state) for p in pilots])
            return True

        if not isinstance(pilots, list):
            pilots = [pilots]

        for pilot in pilots:

            pid   = pilot.uid
            state = pilot.state

            if state in rps.FINAL:

                self._log.debug('pilot %s is final', pid)

                tasks = list()
                for task in self._tasks.values():

                    update = {'uid'             : task.uid,
                              'exception'       : 'RuntimeError("pilot died")',
                              'exception_detail': 'pilot %s is final' % pid,
                              'state'           : rps.FAILED}

                    task._update(update)
                    tasks.append(task.as_dict())

                # final tasks are not pushed
                self.advance(tasks, publish=True, push=False)

        # keep cb registered
        return True


    # --------------------------------------------------------------------------
    #
    def _state_sub_cb(self, topic, msg):

        if self._terminate.is_set():
            return False

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        if cmd != 'update':
            self._log.debug('ignore state cb msg with cmd %s', cmd)
            return True

        things = ru.as_list(arg)
        tasks  = [thing for thing in things if thing.get('type') == 'task']

        self._update_tasks(tasks)

        return True


    # --------------------------------------------------------------------------
    #
    def _update_tasks(self, task_dicts):

        # return information about needed callback and advance activities, so
        # that we don't break bulks here.
        # note however that individual task callbacks are still being called on
        # each task (if any are registered), which can lead to arbitrary,
        # application defined delays.

        to_notify = list()

        with self._tasks_lock:

            for task_dict in task_dicts:

                self._log.debug('task update: %s: %s [%s]',
                                task_dict['uid'], task_dict['state'],
                                task_dict.get('target_state'))

                uid = task_dict['uid']

                # we don't care about tasks we don't know
                task = self._tasks.get(uid)
                if not task:
                    self._log.debug('tmgr: task unknown: %s', uid)
                    continue

                # only update on state changes
                current = task.state
                target  = task_dict['state']

              # if 'target_state' in task_dict:
              #     target = task_dict['target_state']

                if current == target:
                    self._log.debug('tmgr: state known: %s', uid)
                    continue

                target, passed = rps._task_state_progress(uid, current, target)

                if target in [rps.CANCELED, rps.FAILED]:
                    # don't replay intermediate states
                    passed = passed[-1:]

                for s in passed:

                    task_dict['state'] = s
                    self._tasks[uid]._update(task_dict)

                    to_notify.append([task, s])

                task_dict['state'] = target
                ru.dict_merge(self._task_info[uid], task_dict, ru.OVERWRITE)

        if to_notify:
            if _USE_BULK_CB:
                self._bulk_cbs(set([task for task,_ in to_notify]))
            else:
                for task, state in to_notify:
                    self._task_cb(task, state)


    # --------------------------------------------------------------------------
    #
    def _task_cb(self, task, state):

        with self._tcb_lock:

            uid      = task.uid
            cb_dicts = list()
            metric   = rpc.TASK_STATE

            # get wildcard callbacks
            cb_dicts += self._callbacks[metric].get('*', {}).values()
            cb_dicts += self._callbacks[metric].get(uid, {}).values()

            for cb_dict in cb_dicts:

                cb      = cb_dict['cb']
                cb_data = cb_dict['cb_data']

                try:
                    if cb_data: cb(task, state, cb_data)
                    else      : cb(task, state)
                except:
                    self._log.exception('cb error (%s)', cb.__name__)


    # --------------------------------------------------------------------------
    #
    def _bulk_cbs(self, tasks,  metrics=None):

        if not metrics: metrics = [rpc.TASK_STATE]
        else          : metrics = ru.as_list(metrics)

        cbs = dict()  # bulked callbacks to call

        with self._tcb_lock:

            for metric in metrics:

                # get wildcard callbacks
                cb_dicts = self._callbacks[metric].get('*')
                for cb_name in cb_dicts:
                    cbs[cb_name] = {'cb'     : cb_dicts[cb_name]['cb'],
                                    'cb_data': cb_dicts[cb_name]['cb_data'],
                                    'tasks'  : set(tasks)}

                # add task specific callbacks if needed
                for task in tasks:

                    uid = task.uid
                    if uid not in self._callbacks[metric]:
                        continue

                    cb_dicts = self._callbacks[metric].get(uid, {})
                    for cb_name in cb_dicts:

                        if cb_name in cbs:
                            cbs[cb_name]['tasks'].add(task)
                        else:
                            cbs[cb_name] = {'cb'     : cb_dicts[cb_name]['cb'],
                                            'cb_data': cb_dicts[cb_name]['cb_data'],
                                            'tasks'  : set([task])}

            for cb_name in cbs:

                cb      = cbs[cb_name]['cb']
                cb_data = cbs[cb_name]['cb_data']
                objs    = cbs[cb_name]['tasks']

                if cb_data: cb(list(objs), cb_data)
                else      : cb(list(objs))


    # --------------------------------------------------------------------------
    #
    # FIXME: this needs to go to the scheduler
    def _default_wait_queue_size_cb(self, tmgr, wait_queue_size):

        # FIXME: this needs to come from the scheduler?
        if self._terminate.is_set():
            return False

        self._log.info("[Callback]: wait_queue_size: %s.", wait_queue_size)


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        """str: The unique id."""
        return self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def scheduler(self):
        """str: The scheduler name."""

        return self._cfg.get('scheduler')



    # --------------------------------------------------------------------------
    #
    def add_pilots(self, pilots):
        """Associates one or more pilots with the task manager.

        Arguments:
            pilots (radical.pilot.Pilot | list[radical.pilot.Pilot]):
                The pilot objects that will be added to the task manager.

        """

        if not isinstance(pilots, list):
            pilots = [pilots]

        if len(pilots) == 0:
            raise ValueError('cannot add no pilots')

        pilot_docs = list()
        with self._pilots_lock:

            # sanity check, and keep pilots around for inspection
            for pilot in pilots:

                if isinstance(pilot, dict):
                    pilot_dict = pilot

                else:
                    # let pilot know we own it and subscribe for state updates
                    # FIXME: this is not working for pilot dicts (ENTK)
                    pilot.attach_tmgr(self)

                    pilot_dict = pilot.as_dict()
                    pilot.register_callback(self._pilot_state_cb)

                pid = pilot_dict['uid']

                if pid in self._pilots:
                    raise ValueError('pilot %s already added' % pid)
                self._pilots[pid] = pilot
                pilot_docs.append(pilot_dict)

        # publish to the command channel for the scheduler to pick up
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'add_pilots',
                                          'arg' : {'pilots': pilot_docs,
                                                   'tmgr'  : self.uid}})


    # --------------------------------------------------------------------------
    #
    def list_pilots(self):
        """Lists the UIDs of the pilots currently associated with the task manager.

        Returns:
            list[str]: A list of :class:`radical.pilot.Pilot` UIDs.

        """

        with self._pilots_lock:
            return list(self._pilots.keys())


    # --------------------------------------------------------------------------
    #
    def get_pilots(self):
        """Get the pilot instances currently associated with the task manager.

        Returns:
            list[radical.pilot.Pilot]: A list of :class:`radical.pilot.Pilot` instances.

        """

        with self._pilots_lock:
            return list(self._pilots.values())


    # --------------------------------------------------------------------------
    #
    def remove_pilots(self, pilot_ids, drain=False):
        """Disassociates one or more pilots from the task manager.

        After a pilot has been removed from a task manager, it won't process
        any of the task manager's tasks anymore. Calling `remove_pilots`
        doesn't stop the pilot itself.

        Arguments:
            drain (bool): Drain determines what happens to the tasks
                which are managed by the removed pilot(s). If `True`, all tasks
                currently assigned to the pilot are allowed to finish execution.
                If `False` (the default), then non-final tasks will be canceled.

        """

        # TODO: Implement 'drain'.
        # NOTE: the actual removal of pilots from the scheduler is asynchron!

        if drain:
            raise RuntimeError("'drain' is not yet implemented")

        if not isinstance(pilot_ids, list):
            pilot_ids = [pilot_ids]

        if len(pilot_ids) == 0:
            raise ValueError('cannot remove no pilots')

        with self._pilots_lock:

            # sanity check, and keep pilots around for inspection
            for pid in pilot_ids:
                if pid not in self._pilots:
                    raise ValueError('pilot %s not removed' % pid)
                del self._pilots[pid]

        # publish to the control channel for the scheduler to pick up
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'remove_pilots',
                                          'arg' : {'pids'  : pilot_ids,
                                                   'tmgr'  : self.uid}})


    # --------------------------------------------------------------------------
    #
    # FIXME RPC
    def _control_cb(self, topic, msg):

        self._log.debug_5('control cb: %s', msg)

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        if cmd == 'rpc_res':

            self._log.debug('rpc res: %s', arg)
            self._rpc_queue.put(arg)


        elif cmd == 'service_up':

            self._log.debug('service up: %s', arg)
            self._service_update(arg)


    # --------------------------------------------------------------------------
    #
    def _service_update(self, arg):

        uid  = arg['uid']
        task = self._tasks.get(uid)

        if not task:
            return

        task._set_info(arg['info'] or '')  # `None` would indicate an error
        self._log.debug('service update: %s: %s', uid, task.info)


    # --------------------------------------------------------------------------
    #
    def pilot_rpc(self, pid, cmd, *args, rpc_addr=None, **kwargs):
        '''Remote procedure call.

        Send an RPC command and arguments to the pilot and wait for the
        response.  This is a synchronous operation at this point, and it is not
        thread safe to have multiple concurrent RPC calls.
        '''

        if pid not in self._pilots:
            raise ValueError('tmgr does not know pilot %s' % pid)

        return self._pilots[pid].rpc(cmd, *args, rpc_addr=rpc_addr, **kwargs)


    # --------------------------------------------------------------------------
    #
    def list_tasks(self):
        """Get the UIDs of the tasks managed by this task manager.

        Returns:
            list[str]: A list of :class:`radical.pilot.Task` UIDs.

        """

        with self._pilots_lock:
            return list(self._tasks.keys())


    # --------------------------------------------------------------------------
    #
    def submit_raptors(self, descriptions, pilot_id=None):
        """Submit RAPTOR master tasks.

        Submits on or more :class:`radical.pilot.TaskDescription` instances to
        the task manager, where the `TaskDescriptions` have the mode
        `RAPTOR_MASTER` set.

        This is a thin wrapper around `submit_tasks`.

        Arguments:
            descriptions: (radical.pilot.TaskDescription) description of the
                workers to submit.

        Returns:
            list[radical.pilot.Task]: A list of :class:`radical.pilot.Task`
                objects.
        """

        descriptions = ru.as_list(descriptions)

        for td in descriptions:

            if not td.mode == RAPTOR_MASTER:
                raise ValueError('unexpected task mode [%s]' % td.mode)

            raptor_file  = td.get('raptor_file')  or  ''
            raptor_class = td.get('raptor_class') or  'Master'

            td.pilot     = pilot_id
            td.arguments = [raptor_file, raptor_class]

            td.environment['PYTHONUNBUFFERED'] = '1'

            if not td.get('uid'):
                td.uid = ru.generate_id('raptor.%(item_counter)04d',
                                        ru.ID_CUSTOM, ns=self._session.uid)

            if not td.get('executable'):
                td.executable = 'radical-pilot-raptor-master'

            if not td.get('named_env'):
                td.named_env = 'rp'

            # ensure that defaults and backward compatibility kick in
            td.verify()

        return self.submit_tasks(descriptions)


    # --------------------------------------------------------------------------
    #
    def submit_workers(self, descriptions):
        """Submit RAPTOR workers.

        Submits on or more :class:`radical.pilot.TaskDescription` instances to
        the task manager, where the `TaskDescriptions` have the mode
        `RAPTOR_WORKER` set.

        This method is a thin wrapper around `submit_tasks`.

        Arguments:
            descriptions: (radical.pilot.TaskDescription) description of the
                workers to submit.

        Returns:
            list[radical.pilot.Task]: A list of :class:`radical.pilot.Task`
                objects.
        """

        descriptions = ru.as_list(descriptions)

        for td in descriptions:

            if not td.mode == RAPTOR_WORKER:
                raise ValueError('unexpected task mode [%s]' % td.mode)

            raptor_id    = td.get('raptor_id')
            raptor_file  = td.get('raptor_file')  or  ''
            raptor_class = td.get('raptor_class') or  'DefaultWorker'

            if not raptor_id:
                raise ValueError('RAPTOR_WORKER descriptions need `raptor_id`')

            if not td.get('uid'):
                td.uid = ru.generate_id(raptor_id + '.%(item_counter)04d',
                                        ru.ID_CUSTOM, ns=self._session.uid)

            if not td.get('executable'):
                td.executable = 'radical-pilot-raptor-worker'

            if not td.get('named_env'):
                td.named_env = 'rp'

            td.raptor_id = raptor_id
            td.arguments = [raptor_file, raptor_class, raptor_id]

            td.environment['PYTHONUNBUFFERED'] = '1'

            # ensure that defaults and backward compatibility kick in
            td.verify()

        return self.submit_tasks(descriptions)


    # --------------------------------------------------------------------------
    #
    def submit_tasks(self, descriptions):
        """Submit tasks for execution.

        Submits one or more :class:`radical.pilot.Task` instances to the task
        manager.

        Arguments:
            descriptions (radical.pilot.TaskDescription | list
                [radical.pilot.TaskDescription]):
                The description of the task instance(s) to create.

        Returns:
            list[radical.pilot.Task]: A list of :class:`radical.pilot.Task`
                objects.

        """

        from .task import Task

        if not descriptions:
            return []

        ret_list = True
        if descriptions and not isinstance(descriptions, list):
            ret_list     = False
            descriptions = [descriptions]

        # we return a list of tasks
        tasks = list()
        ret   = list()
        self._rep.progress_tgt(len(descriptions), label='submit')
        for td in descriptions:

            mode = td.mode

            if mode == RAPTOR_MASTER:
                task = RaptorMaster(tmgr=self, descr=td, origin='client')

            elif mode == RAPTOR_WORKER:
                task = RaptorWorker(tmgr=self, descr=td, origin='client')

            else:
                task = Task(tmgr=self, descr=td, origin='client')

            tasks.append(task)
            self._rep.progress()

            if len(tasks) >= 1024:
                # submit this bulk
                with self._tasks_lock:
                    for task in tasks:
                        self._tasks[task.uid] = task

                task_docs = [u.as_dict() for u in tasks]
                self.advance(task_docs, rps.TMGR_SCHEDULING_PENDING,
                             publish=True, push=True)
                ret += tasks
                tasks = list()

        # submit remaining bulk (if any)
        if tasks:
            with self._tasks_lock:
                for task in tasks:
                    self._tasks[task.uid] = task

            task_docs = [t.as_dict() for t in tasks]
            self.advance(task_docs, rps.TMGR_SCHEDULING_PENDING,
                         publish=True, push=True)
            ret += tasks

        self._rep.progress_done()

        if ret_list: return ret
        else       : return ret[0]


    # --------------------------------------------------------------------------
    #
    def get_tasks(self, uids=None):
        """Get one or more tasks identified by their IDs.

        Arguments:
            uids (str | list[str]): The IDs of the task objects to return.

        Returns:
            list[radical.pilot.Task]:
                A list of :class:`radical.pilot.Task` objects.

        """

        if not uids:
            with self._tasks_lock:
                ret = list(self._tasks.values())
            return ret

        ret_list = True
        if (not isinstance(uids, list)) and (uids is not None):
            ret_list = False
            uids = [uids]

        ret = list()
        with self._tasks_lock:
            for uid in uids:
                if uid not in self._tasks:
                    raise ValueError('task %s not known' % uid)
                ret.append(self._tasks[uid])

        if ret_list: return ret
        else       : return ret[0]


    # --------------------------------------------------------------------------
    #
    def wait_tasks(self, uids=None, state=None, timeout=None):
        """Block for state transition.

        Returns when one or more :class:`radical.pilot.Tasks` reach a
        specific state.

        If `uids` is `None`, `wait_tasks` returns when **all**
        Tasks reach the state defined in `state`.  This may include
        tasks which have previously terminated or waited upon.

        Example::

            # TODO -- add example

        Arguments:
            uids (str | list[str]): If uids is set, only the Tasks with the
                specified uids are considered. If uids is `None` (default), all
                Tasks are considered.
            state (str): The state that Tasks have to reach in order for the call
                to return.

                By default `wait_tasks` waits for the Tasks to
                reach a terminal state, which can be one of the following.

                * :data:`radical.pilot.rps.DONE`
                * :data:`radical.pilot.rps.FAILED`
                * :data:`radical.pilot.rps.CANCELED`
            timeout (float):
                Timeout in seconds before the call returns regardless of Pilot
                state changes. The default value **None** waits forever.

        """

        ret_list = True
        if not uids:
            with self._tasks_lock:
                uids = self._tasks.keys()
        else:
            if not isinstance(uids, list):
                ret_list = False
                uids = [uids]

        if   not state                  : states = rps.FINAL
        elif not isinstance(state, list): states = [state]
        else                            : states =  state

        self._log.debug('wait for %s: %s', uids, states)

        # we simplify state check by waiting for the *earliest* of the given
        # states - if the task happens to be in any later state, we are sure the
        # earliest has passed as well.
        check_state_val = rps._task_state_values[rps.FINAL[-1]]
        for state in states:
            check_state_val = min(check_state_val,
                                  rps._task_state_values[state])

        start    = time.time()
        to_check = None

        with self._tasks_lock:
            to_check = [self._tasks[uid] for uid in uids]

        # We don't want to iterate over all tasks again and again, as that would
        # duplicate checks on tasks which were found in matching states.  So we
        # create a list from which we drop the tasks as we find them in
        # a matching state
        self._rep.progress_tgt(len(to_check), label='wait')
        self._log.debug('WAIT for %d tasks', len(to_check))
        while to_check and not self._terminate.is_set():

          # self._log.debug('wait for %d tasks', len(to_check))

            # check timeout
            if timeout and (timeout <= (time.time() - start)):
                self._log.debug ("wait timed out")
                break

            time.sleep (0.1)

            # FIXME: print percentage...
          # print 'wait tasks: %s' % [[u.uid, u.state] for u in to_check]

            check_again = list()
            for task in to_check:

                # we actually don't check if a task is in a specific (set of)
                # state(s), but rather check if it ever *has been* in any of
                # those states
                if task.state not in rps.FINAL and \
                    rps._task_state_values[task.state] < check_state_val:

                    # this task does not match the wait criteria
                  # self._log.debug('wait again for %s [%s]', task.uid, task.state)
                    check_again.append(task)

                else:
                    # stop watching this task
                  # self._log.debug('wait ok    for %s [%s]', task.uid, task.state)
                    if task.state in [rps.FAILED]:
                        self._rep.progress()  # (color='error', c='-')
                    elif task.state in [rps.CANCELED]:
                        self._rep.progress()  # (color='warn', c='*')
                    else:
                        self._rep.progress()  # (color='ok', c='+')

            to_check = check_again

        self._log.debug('wait completed')

        self._rep.progress_done()


        # grab the current states to return
        state = None
        with self._tasks_lock:
            states = [self._tasks[uid].state for uid in uids]

        sdict = {state: states.count(state) for state in set(states)}
        for state in sorted(set(states)):
            self._rep.info('\t%-10s: %5d\n' % (state, sdict[state]))

        if to_check: self._rep.warn('>>timeout\n')
        else       : self._rep.ok  ('>>ok\n')

        # done waiting
        if ret_list: return states
        else       : return states[0]


    # --------------------------------------------------------------------------
    #
    def cancel_tasks(self, uids=None):
        """Cancel one or more :class:`radical.pilot.Task` s.

        Note that cancellation of tasks is not immediate, i.e. their state is
        not necessarily `CANCELED` after this method returns.  Instead, the
        cancellation request is sent to the components which currently manage
        the tasks and which then will enact the request at their discretion,
        eventually leading to the state transition to `CANCELLED`.

        Arguments:
            uids (str | list[str]): The IDs of the tasks to cancel.
        """

        if not uids:
            with self._tasks_lock:
                uids  = list(self._tasks.keys())
        else:
            if not isinstance(uids, list):
                uids = [uids]

        # we *always* issue the cancellation command to the local components
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'cancel_tasks',
                                          'arg' : {'uids' : uids,
                                                   'tmgr' : self.uid},
                                          'fwd' : True})

        # We do not wait and block the call until all the tasks are marked
        # cancelled.  This means when inspecting for state just after a state
        # change, we may observe a old state, instead of CANCELLED.


    # --------------------------------------------------------------------------
    #
    # TODO: `metric` -> `metrics`, for consistency with `unregister_callback()`
    #
    def register_callback(self, cb, cb_data=None, metric=None, uid=None):
        """Registers a new callback function with the TaskManager.

        Manager-level
        callbacks get called if the specified metric changes.  The default
        metric `TASK_STATE` fires the callback if any of the Tasks
        managed by the PilotManager change their state.

        All callback functions need to have the same signature::

            def cb(obj, value) -> None:
                ...

        where ``obj`` is a handle to the object that triggered the callback,
        ``value`` is the metric, and ``data`` is the data provided on
        callback registration..  In the example of `TASK_STATE` above, the
        object would be the task in question, and the value would be the new
        state of the task.

        If ``cb_data`` is given, then the ``cb`` signature changes to
        ::

            def cb(obj, state, cb_data) -> None:
                ...

        and ``cb_data`` are passed unchanged.

        If ``uid`` is given, the callback will invoked only for the specified
        task.

        Available metrics are

        * `TASK_STATE`: fires when the state of any of the tasks which are
          managed by this task manager instance is changing.  It communicates
          the task object instance and the tasks new state.
        * `WAIT_QUEUE_SIZE`: fires when the number of unscheduled tasks (i.e.
          of tasks which have not been assigned to a pilot for execution)
          changes.

        """

        # FIXME: the signature should be (self, metrics, cb, cb_data)

        if not metric:
            metric = rpc.TASK_STATE

        metrics = ru.as_list(metric)

        if not uid:
            uid = '*'

        elif uid not in self._tasks:
            raise ValueError('no such task %s' % uid)


        with self._tcb_lock:

            for metric in metrics:

                if metric not in rpc.TMGR_METRICS:
                    raise ValueError("invalid tmgr metric '%s'" % metric)

                cb_id = id(cb)

                if metric not in self._callbacks:
                    self._callbacks[metric] = dict()

                if uid not in self._callbacks[metric]:
                    self._callbacks[metric][uid] = dict()

                self._callbacks[metric][uid][cb_id] = {'cb'     : cb,
                                                       'cb_data': cb_data}


    # --------------------------------------------------------------------------
    #
    def unregister_callback(self, cb=None, metrics=None, uid=None):

        if not metrics:
            metrics = rpc.TASK_STATE

        metrics = ru.as_list(metrics)

        if not uid:
            uid = '*'

        elif uid not in self._tasks:
            raise ValueError('no such task %s' % uid)

        for metric in metrics:
            if metric not in rpc.TMGR_METRICS:
                raise ValueError("invalid tmgr metric '%s'" % metric)

        with self._tcb_lock:

            for metric in metrics:

                if metric not in self._callbacks:
                    raise ValueError("cb metric '%s' invalid" % metric)

                if uid not in self._callbacks[metric]:
                    raise ValueError("cb target '%s' invalid" % uid)

                if cb:
                    to_delete = [id(cb)]
                else:
                    to_delete = list(self._callbacks[metric][uid].keys())

                for cb_id in to_delete:

                    if cb_id not in self._callbacks[metric][uid]:
                        raise ValueError("cb %s not registered" % cb_id)

                    del self._callbacks[metric][uid][cb_id]


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import time
import threading as mt

import radical.utils as ru

from .  import utils     as rpu
from .  import states    as rps
from .  import constants as rpc

from .staging_directives import expand_staging_directives


# bulk callbacks are implemented, but are currently not used nor exposed.
_USE_BULK_CB = False
if os.environ.get('RADICAL_PILOT_BULK_CB', '').lower() in ['true', 'yes', '1']:
    _USE_BULK_CB = True


# ------------------------------------------------------------------------------
#
class PilotManager(rpu.ClientComponent):
    """Manage Pilot instances.

    A PilotManager manages :class:`rp.Pilot` instances that are
    submitted via the :func:`radical.pilot.PilotManager.submit_pilots` method.
    It is possible to attach one or more
    :ref:`HPC resources </tutorials/configuration.ipynb#Platform-description>`
    to a PilotManager to outsource machine specific configuration parameters
    to an external configuration file.

    Example::

        s = rp.Session()

        pm = rp.PilotManager(session=s)

        pd = rp.PilotDescription()
        pd.resource = "futuregrid.alamo"
        pd.cpus = 16

        p1 = pm.submit_pilots(pd)  # create first  pilot with 16 cores
        p2 = pm.submit_pilots(pd)  # create second pilot with 16 cores

        # Create a workload of 128 '/bin/sleep' tasks
        tasks = []
        for task_count in range(0, 128):
            t = rp.TaskDescription()
            t.executable = "/bin/sleep"
            t.arguments = ['60']
            tasks.append(t)

        # Combine the two pilots, the workload and a scheduler via
        # a TaskManager.
        tm = rp.TaskManager(session=session, scheduler=rp.SCHEDULER_ROUND_ROBIN)
        tm.add_pilot(p1)
        tm.submit_tasks(tasks)


    The pilot manager can issue notification on pilot state changes.  Whenever
    state notification arrives, any callback registered for that notification is
    fired.

    Note:
        State notifications can arrive out of order wrt the pilot state model!

    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, session, cfg='default'):
        """Creates a new PilotManager and attaches is to the session.

        Arguments:
            session (rp.Session): The session instance to use.
            uid (str): ID for pilot manager, to be used for reconnect
            cfg (dict, str): The configuration or name of configuration to use.

        Returns:
            rp.PilotManager: A new `PilotManager` object.

        """

        assert session._role == session._PRIMARY, 'pmgr needs primary session'

        # initialize the base class (with no intent to fork)
        self._uid         = ru.generate_id('pmgr.%(item_counter)04d',
                                            ru.ID_CUSTOM, ns=session.uid)

        self._uids        = list()   # known UIDs
        self._pilots      = dict()
        self._pilots_lock = mt.RLock()
        self._callbacks   = dict()
        self._pcb_lock    = mt.RLock()
        self._terminate   = mt.Event()
        self._closed      = False

        for m in rpc.PMGR_METRICS:
            self._callbacks[m] = dict()

        # NOTE: `name` and `cfg` are overloaded, the user cannot point to
        #       a predefined config and amed it at the same time.  This might
        #       be ok for the session, but introduces a minor API inconsistency.
        #
        name = None
        if isinstance(cfg, str):
            name = cfg
            cfg  = None

        cfg                = ru.Config('radical.pilot.pmgr', name=name, cfg=cfg)
        cfg.uid            = self._uid
        cfg.owner          = self._uid
        cfg.sid            = session.uid
        cfg.path           = session.path
        cfg.reg_addr       = session.reg_addr
        cfg.heartbeat      = session.cfg.heartbeat
        cfg.client_sandbox = session._get_client_sandbox()

        super().__init__(cfg, session=session)
        self.start()

        self._log.info('started pmgr %s', self._uid)

        self._rep = self._session._get_reporter(name=self._uid)
        self._rep.info('<<create pilot manager')

        # create pmgr bridges and components, use session cmgr for that
        self._cmgr = rpu.ComponentManager(cfg.sid, cfg.reg_addr, self._uid)
        self._cmgr.start_bridges(self._cfg.bridges)
        self._cmgr.start_components(self._cfg.components)

        self._session._register_pmgr(self)

        # The output queue is used to forward submitted pilots to the
        # launching component.
        self.register_output(rps.PMGR_LAUNCHING_PENDING,
                             rpc.PMGR_LAUNCHING_QUEUE)

        self._stager = rpu.StagingHelper(self._log)

        # register the state notification pull cb and hb pull cb
        # FIXME: we may want to have the frequency configurable
        # FIXME: this should be a tailing cursor in the update worker
        self.register_timed_cb(self._pilot_heartbeat_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # also listen to the state pubsub for pilot state changes
        self.register_subscriber(rpc.STATE_PUBSUB, self._state_sub_cb)

        # let session know we exist
        self._session._register_pmgr(self)

        self._prof.prof('setup_done', uid=self._uid)
        self._rep.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # the manager must not carry bridge and component handles across forks
        ru.atfork(self._atfork_prepare, self._atfork_parent, self._atfork_child)


    # --------------------------------------------------------------------------
    #
    # EnTK forks, make sure we don't carry traces of children across the fork
    #
    def _atfork_prepare(self): pass
    def _atfork_parent(self) : pass
    def _atfork_child(self)  :
        self._bridges    = dict()
        self._components = dict()


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self._cmgr.close()
        self._fail_missing_pilots()


    # --------------------------------------------------------------------------
    #
    def close(self, terminate=True):
        """Shut down the PilotManager and all its components.

        Arguments:
            terminate (bool): cancel non-final pilots if True (default)

                Note:
                    Pilots cannot be reconnected to after termination.

        """

        if self._closed:
            return

        self._rep.info('<<close pilot manager')

        # disable callbacks during shutdown
        # FIXME: really?
        with self._pcb_lock:
            for m in rpc.PMGR_METRICS:
                self._callbacks[m] = dict()

        # If terminate is set, kill all pilots.
        if terminate:

            # skip reporting for `wait_pilots`
            is_rep_enabled = self._rep._enabled
            self._rep._enabled = False
            self.cancel_pilots(_timeout=1)
            self._rep._enabled = is_rep_enabled

            self.kill_pilots(_timeout=10)

        self._terminate.set()
        self._cmgr.close()

        self._log.info("Closed PilotManager %s." % self._uid)

        self._closed = True
        self._rep.ok('>>ok\n')

        self.dump()

        super().close()


    # --------------------------------------------------------------------------
    #
    def dump(self, name=None):

        # dump json
        json = self.as_dict()
        json['type']   = 'pmgr'
        json['uid']    = self.uid
        json['pilots'] = [pilot.as_dict() for pilot in self._pilots.values()]

        if name:
            tgt = '%s/%s.%s.json' % (self._session.path, self.uid, name)
        else:
            tgt = '%s/%s.json' % (self._session.path, self.uid)

        ru.write_json(json, tgt)

    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a dictionary representation of the PilotManager object."""

        ret = {
            'uid': self.uid,
            'cfg': self.cfg
        }

        return ret


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        """Returns a string representation of the PilotManager object."""

        return str(self.as_dict())


    # --------------------------------------------------------------------------
    #
    def _pilot_heartbeat_cb(self):

        if self._terminate.is_set():
            return False

        # send heartbeat
        self._pilot_send_hb()
        return True


    # --------------------------------------------------------------------------
    #
    def _state_sub_cb(self, topic, msg):

        if self._terminate.is_set():
            return False

        self._log.debug('state event: %s', msg)

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        if cmd != 'update':
            self._log.debug('ignore state cb msg with cmd %s', cmd)
            return True

        if isinstance(arg, list): things =  arg
        else                    : things = [arg]

        for thing in things:

            if 'type' in thing and thing['type'] == 'pilot':

                self._log.debug('state push: %s: %s', thing['uid'],
                                thing['state'])

                # we got the state update from the state callback - don't
                # publish it again
                if not self._update_pilot(thing, publish=False):
                    return False

        return True


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):

        if self._terminate.is_set():
            return False

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        self._log.debug_9('got control cmd %s: %s', cmd, arg)

        if cmd == 'pilot_activate':
            pilot = arg['pilot']
            self._update_pilot(pilot, publish=True)

            # store resource json for RA
            fname = '%s/%s.resources.json' % (self._cfg.path, pilot['uid'])
            ru.write_json(fname, pilot['resources'])


    # --------------------------------------------------------------------------
    #
    def _update_pilot(self, pilot_dict, publish=False, advance=True):

        # FIXME: this is breaking the bulk!

        pid   = pilot_dict['uid']
      # state = pilot_dict['state']

        with self._pilots_lock:

            # we don't care about pilots we don't know
            if pid not in self._pilots:
                return   # this is not an error

            # only update on state changes
            current = self._pilots[pid].state
            target  = pilot_dict['state']

            # always update the pilot instance, even if state didn't change
            if current == target:
                self._pilots[pid]._update(pilot_dict)
                return

            target, passed = rps._pilot_state_progress(pid, current, target)
          # self._log.debug('%s current: %s', pid, current)
          # self._log.debug('%s target : %s', pid, target )
          # self._log.debug('%s passed : %s', pid, passed )

            if target in [rps.CANCELED, rps.FAILED]:
                # don't replay intermediate states
                passed = passed[-1:]

            for s in passed:
                self._log.debug('%s advance: %s', pid, s )
                # we got state from either pubsub or DB, so don't publish again.
                # we also don't need to maintain bulks for that reason.
                pilot_dict['state'] = s
                self._pilots[pid]._update(pilot_dict)

                if advance:
                    self.advance(pilot_dict, s, publish=publish, push=False)

                if s in [rps.PMGR_ACTIVE]:
                    self._log.info('pilot %s is %s: %s [%s]', pid, s,
                                    pilot_dict.get('lm_info'),
                                    pilot_dict.get('lm_detail'))


    # --------------------------------------------------------------------------
    #
    def _call_pilot_callbacks(self, pilot):

        state = pilot.state

        with self._pcb_lock:

            for cb_dict in self._callbacks[rpc.PILOT_STATE].values():

                cb      = cb_dict['cb']
                cb_data = cb_dict['cb_data']

              # print ' ~~~ call PCBS: %s -> %s : %s [%s]' \
              #         % (self.uid, pilot.state, cb_name, cb_data)
                self._log.debug('pmgr calls cb %s for %s', pilot.uid, cb)

                if _USE_BULK_CB:
                    if cb_data: cb([pilot], cb_data)
                    else      : cb([pilot])
                else:
                    if cb_data: cb(pilot, state, cb_data)
                    else      : cb(pilot, state)


    # --------------------------------------------------------------------------
    #
    def _pilot_send_hb(self, pid=None):

        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'pmgr_heartbeat',
                                          'arg' : {'pmgr' : self.uid}})


    # --------------------------------------------------------------------------
    #
    def _pilot_staging_input(self, pid, sds):
        """Run some staging directives for a pilot."""

        for sd in expand_staging_directives(sds):
            self._prof.prof('staging_in_start', uid=pid, msg=sd['uid'])
            self._stager.handle_staging_directive(sd)
            self._prof.prof('staging_in_stop', uid=pid, msg=sd['uid'])


    # --------------------------------------------------------------------------
    #
    def _pilot_staging_output(self, pid, sds):
        """Run some staging directives for a pilot."""

        for sd in expand_staging_directives(sds):
            self._prof.prof('staging_out_start', uid=self.uid, msg=sd['uid'])
            self._stager.handle_staging_directive(sd)
            self._prof.prof('staging_out_stop', uid=self.uid, msg=sd['uid'])


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        """str: The unique id."""
        return self._uid


    # --------------------------------------------------------------------------
    #
    def list_pilots(self):
        """Get the UIDs of the managed :class:`rp.Pilots`.

        Returns:
            list[str]: A list of :class:`rp.Pilot` UIDs.

        """

        with self._pilots_lock:
            ret = list(self._pilots.keys())

        return ret


    # --------------------------------------------------------------------------
    #
    def submit_pilots(self, descriptions):
        """Submit one or more `rp.Pilot` instances to the pilot manager.

        Arguments:
            descriptions (rp.PilotDescription | list[rp.PilotDescription]):
                The description of the pilot instance(s) to create.

        Returns:
            list[rp.Pilot]: A list of :class:`rp.Pilot` objects.

        """

        from .pilot import Pilot

        ret_list = True
        if not isinstance(descriptions, list):
            ret_list     = False
            descriptions = [descriptions]

        if len(descriptions) == 0:
            raise ValueError('cannot submit no pilot descriptions')


        self._rep.info('<<submit %d pilot(s)' % len(descriptions))

        # create the pilot instance
        pilots     = list()
        pilot_docs = list()
        for pd in descriptions :

            pilot = Pilot(pmgr=self, descr=pd)
            pilots.append(pilot)
            pilot_doc = pilot.as_dict()
            pilot_docs.append(pilot_doc)

            # keep pilots around
            with self._pilots_lock:
                self._pilots[pilot.uid] = pilot

            if pd.get('nodes'):
                self._rep.plain('\n\t%s   %-20s %6d nodes' %
                                (pilot.uid, pd['resource'], pd['nodes']))
            else:
                self._rep.plain('\n\t%s   %-20s %6d cores  %6d gpus' %
                                (pilot.uid, pd['resource'],
                                 pd.get('cores', 0), pd.get('gpus', 0)))

        # initial state advance to 'NEW'
        # FIXME: we should use update_pilot(), but that will not trigger an
        #        advance, since the state did not change.  We would then miss
        #        the profile entry for the advance to NEW.  So we here basically
        #        only trigger the profile entry for NEW.
        self.advance(pilot_docs, state=rps.NEW, publish=False, push=False)

        # immediately send first heartbeat
        for pilot_doc in pilot_docs:
            pid = pilot_doc['uid']
            self._pilot_send_hb(pid)

        # Only after the insert/update can we hand the pilots over to the next
        # components (ie. advance state).
        for pd in pilot_docs:
            pd['state'] = rps.PMGR_LAUNCHING_PENDING
            self._update_pilot(pd, advance=False)
        self.advance(pilot_docs, publish=True, push=True)


        self._rep.ok('>>ok\n')

        if ret_list: return pilots
        else       : return pilots[0]


    # --------------------------------------------------------------------------
    #
    def _reconnect_pilots(self):
        """Restore Pilot info from database.

        When reconnecting, we need to dig information about all pilots from the
        DB for which this pmgr is responsible.
        """

        from .pilot             import Pilot
        from .pilot_description import PilotDescription

      # self.is_valid()

        # FIXME MONGODB
        # pilot_docs = self._session._dbs.get_pilots(pmgr_uid=self.uid)

        # with self._pilots_lock:
        #     for ud in pilot_docs:

        #         descr = PilotDescription(ud['description'])
        #         pilot = Pilot(pmgr=self, descr=descr)

        #         self._pilots[pilot.uid] = pilot


    # --------------------------------------------------------------------------
    #
    def get_pilots(self, uids=None):
        """Returns one or more pilots identified by their IDs.

        Arguments:
            uids (str | list[str]): The IDs of the pilot objects to return.

        Returns:
            list: A list of :class:`rp.Pilot` objects.

        """

        if not uids:
            with self._pilots_lock:
                ret = list(self._pilots.values())
            return ret


        ret_list = True
        if (not isinstance(uids, list)) and (uids is not None):
            ret_list = False
            uids = [uids]

        ret = list()
        with self._pilots_lock:
            for uid in uids:
                if uid not in self._pilots:
                    raise ValueError('pilot %s not known' % uid)
                ret.append(self._pilots[uid])

        if ret_list: return ret
        else       : return ret[0]


    # --------------------------------------------------------------------------
    #
    def wait_pilots(self, uids=None, state=None, timeout=None):
        """Block for state transition.

        Returns when one or more :class:`rp.Pilots` reach a
        specific state.

        If `pilot_uids` is `None`, `wait_pilots` returns when **all**
        Pilots reach the state defined in `state`.  This may include
        pilots which have previously terminated or waited upon.

        Example:
            # TODO -- add example

        Arguments:
            uids (str | list[str], optional): If set, only the Pilots with the
                specified uids are considered. If `None` (default), all
                Pilots are considered.
            state (str | list[str]): The state that Pilots have to reach in order
                for the call to return.

                By default `wait_pilots` waits for the Pilots to
                reach a terminal state, which can be one of the following:
                * :data:`rp.rps.DONE`
                * :data:`rp.rps.FAILED`
                * :data:`rp.rps.CANCELED`
            timeout (float, optional): Timeout in seconds before the call returns
                regardless of Pilot
                state changes. The default value **None** waits forever.

        """

        if not uids:
            with self._pilots_lock:
                uids = list()
                for uid,pilot in self._pilots.items():
                    if pilot.state not in rps.FINAL:
                        uids.append(uid)

        if not state:
            states = rps.FINAL
        elif isinstance(state, list):
            states = state
        else:
            states = [state]

        ret_list = True
        if not isinstance(uids, list):
            ret_list = False
            uids     = [uids]

        self._rep.info('<<wait for %d pilot(s)\n\t' % len(uids))

        start    = time.time()
        to_check = None

        with self._pilots_lock:

            for uid in uids:
                if uid not in self._pilots:
                    raise ValueError('pilot %s not known' % uid)

            to_check = [self._pilots[uid] for uid in uids]

        # We don't want to iterate over all pilots again and again, as that
        # would duplicate checks on pilots which were found in matching states.
        # So we create a list from which we drop the pilots as we find them in
        # a matching state
        self._rep.idle(mode='start')
        while to_check and not self._terminate.is_set():

            self._rep.idle()

            to_check = [pilot for pilot in to_check
                               if pilot.state not in states and
                                  pilot.state not in rps.FINAL]

            if to_check:

                if timeout and (timeout <= (time.time() - start)):
                    self._log.debug ("wait timed out")
                    break

            time.sleep (0.1)

        self._rep.idle(mode='stop')

        if to_check: self._rep.warn('>>timeout\n')
        else       : self._rep.ok  ('>>ok\n')

        # grab the current states to return
        state = None
        with self._pilots_lock:
            states = [self._pilots[uid].state for uid in uids]

        # done waiting
        if ret_list: return states
        else       : return states[0]


    # --------------------------------------------------------------------------
    #
    def _fail_missing_pilots(self):
        """Advance remaining pilots to failed state.

        During termination, fail all pilots for which we did not manage to
        obtain a final state - we trust that they'll follow up on their
        cancellation command in due time, if they can
        """

        pass

      # with self._pilots_lock:
      #     for pid in self._pilots:
      #         pilot = self._pilots[pid]
      #         if pilot.state not in rps.FINAL:
      #             self.advance(pilot.as_dict(), rps.FAILED,
      #                          publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def cancel_pilots(self, uids=None, _timeout=None):
        """Cancel one or more :class:`rp.Pilots`.

        Arguments:
            uids (str | list[str], optional): The IDs of the pilots to cancel.

        """

        if not uids:
            with self._pilots_lock:
                uids = list(self._pilots.keys())

        if not isinstance(uids, list):
            uids = [uids]

        self._log.debug('pilot(s).need(s) cancellation %s', uids)

        # send the cancellation request to the pilots
        self._log.debug('issue cancel_pilots for %s', uids)
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'cancel_pilots',
                                          'arg' : {'pmgr' : self.uid,
                                                   'uids' : uids},
                                          'fwd' : True})
        # wait for the cancel to be enacted
        self.wait_pilots(uids=uids, timeout=_timeout)

        # FIXME: only finalize pilots which actually terminated
        with self._pilots_lock:
            for uid in uids:
                if uid not in self._pilots:
                    raise ValueError('pilot %s not known' % uid)


    # --------------------------------------------------------------------------
    #
    def kill_pilots(self, uids=None, _timeout=None):
        """Kill one or more :class:`rp.Pilots`

        Arguments:
            uids (str | list[str], optional): The IDs of the pilots to cancel.

        """

        if not uids:
            with self._pilots_lock:
                uids = list(self._pilots.keys())

        if not isinstance(uids, list):
            uids = [uids]

        with self._pilots_lock:
            for uid in uids:
                if uid not in self._pilots:
                    raise ValueError('pilot %s not known' % uid)

        self._log.debug('pilot(s).need(s) killing %s', uids)

        # inform pmgr.launcher - it will force-kill the pilots
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'kill_pilots',
                                          'arg' : {'pmgr' : self.uid,
                                                   'uids' : uids}})

        # wait for the kill to be enacted
        self.wait_pilots(uids=uids, timeout=_timeout)


    # --------------------------------------------------------------------------
    #
    def register_callback(self, cb, cb_data=None, metric=rpc.PILOT_STATE):
        """Registers a new callback function with the PilotManager.

        Manager-level callbacks get called if the specified metric changes.
        The default metric `PILOT_STATE` fires the callback if any of the Pilots
        managed by the PilotManager change their state.

        All callback functions need to have the same signature::

            def cb(obj, value, cb_data)

        where ``object`` is a handle to the object that triggered the callback,
        ``value`` is the metric, and ``data`` is the data provided on
        callback registration..  In the example of `PILOT_STATE` above, the
        object would be the pilot in question, and the value would be the new
        state of the pilot.

        Available metrics are

        * `PILOT_STATE`: fires when the state of any of the pilots which are
            managed by this pilot manager instance is changing.  It communicates
            the pilot object instance and the pilots new state.

        """

        # FIXME: the signature should be (self, metrics, cb, cb_data)

        if metric not in rpc.PMGR_METRICS :
            raise ValueError ("invalid pmgr metric '%s'" % metric)

        with self._pcb_lock:
            cb_id = id(cb)
            self._callbacks[metric][cb_id] = {'cb'      : cb,
                                              'cb_data' : cb_data}


    # --------------------------------------------------------------------------
    #
    def unregister_callback(self, cb, metric=rpc.PILOT_STATE):

        if metric and metric not in rpc.PMGR_METRICS :
            raise ValueError ("invalid pmgr metric '%s'" % metric)

        if not metric:
            metrics = rpc.PMGR_METRICS
        elif isinstance(metric, list):
            metrics =  metric
        else:
            metrics = [metric]

        with self._pcb_lock:

            for metric in metrics:

                if cb:
                    to_delete = [id(cb)]
                else:
                    to_delete = list(self._callbacks[metric].keys())

                for cb_id in to_delete:

                    if cb_id not in self._callbacks[metric]:
                        raise ValueError("unknown callback '%s'" % cb_id)

                    del self._callbacks[metric][cb_id]


    # --------------------------------------------------------------------------
    #
    def check_uid(self, uid):

        # ensure that uid is not yet known
        if uid in self._uids:
            return False
        else:
            self._uids.append(uid)
            return True


# ------------------------------------------------------------------------------




__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import copy
import time

import threading     as mt

import radical.utils as ru

from . import states    as rps
from . import utils     as rpu
from . import constants as rpc

from .staging_directives import expand_description
from .task_description   import TaskDescription, TASK_SERVICE
from .resource_config    import Slot


_uids = list()


# ------------------------------------------------------------------------------
#
def _check_uid(uid):
    # ensure that uid is not yet known

    if uid in _uids:
        return False
    else:
        _uids.append(uid)
        return True


# ------------------------------------------------------------------------------
#
class Task(object):
    """Represent a 'task' that is executed on a Pilot.

    Tasks allow to control and query the state of this task.

    Note:
        A task cannot be created directly. The factory method
        :func:`rp.TaskManager.submit_tasks` has to be used instead.

    Example::

        tmgr = rp.TaskManager(session=s)

        ud = rp.TaskDescription()
        ud.executable = "/bin/date"

        task = tmgr.submit_tasks(ud)

    """
    # --------------------------------------------------------------------------
    # In terms of implementation, a Task is not much more than a dict whose
    # content are dynamically updated to reflect the state progression through
    # the TMGR components.  As a Task is always created via a TMGR, it is
    # considered to *belong* to that TMGR, and all activities are actually
    # implemented by that TMGR.
    #
    # Note that this implies that we could create CUs before submitting them
    # to a TMGR, w/o any problems. (FIXME?)
    # --------------------------------------------------------------------------


    # --------------------------------------------------------------------------
    #
    def __init__(self, tmgr, descr, origin):

        # ensure that the description is viable
        descr.verify()

        # 'static' members
        self._tmgr   = tmgr
        self._descr  = descr
        self._origin = origin

        # initialize state
        self._session          = self._tmgr.session
        self._uid              = self._descr.uid
        self._state            = rps.NEW
        self._log              = tmgr._log
        self._exit_code        = None
        self._stdout           = str()
        self._stderr           = str()
        self._ofiles           = None
        self._return_value     = None
        self._exception        = None
        self._exception_detail = None
        self._info             = None
        self._info_evt         = mt.Event()
        self._pilot            = self._descr.pilot
        self._endpoint_fs      = None
        self._resource_sandbox = None
        self._session_sandbox  = None
        self._pilot_sandbox    = None
        self._task_sandbox     = None
        self._client_sandbox   = None
        self._callbacks        = dict()
        self._slots            = None
        self._partition        = None

        # ensure uid is unique
        if self._uid:
            if not _check_uid(self._uid):
                raise ValueError('uid %s is not unique' % self._uid)
        else:
            self._uid = ru.generate_id('task.%(item_counter)06d', ru.ID_CUSTOM,
                                       ns=self._session.uid)

        for m in rpc.TMGR_METRICS:
            self._callbacks[m] = dict()

        # we always invke the default state cb
        self._callbacks[rpc.TASK_STATE][self._default_state_cb.__name__] = {
                'cb'      : self._default_state_cb,
                'cb_data' : None}

        # If staging directives exist, expand them to the full dict version.  Do
        # not, however, expand any URLs as of yet, as we likely don't have
        # sufficient information about pilot sandboxes etc.
        expand_description(self._descr)

        self._tmgr.advance(self.as_dict(), rps.NEW, publish=False, push=False)


    # --------------------------------------------------------------------------
    #
    def __str__(self):

        return str([self.uid, self.pilot, self.state])


    # --------------------------------------------------------------------------
    #
    def _default_state_cb(self, task, state=None):

        self._log.info("[Callback]: task %s state: %s.", self.uid, self.state)


    # --------------------------------------------------------------------------
    #
    def _update(self, task_dict, reconnect=False):
        """State change updater.

        This will update the facade object after state changes etc, and is
        invoked by whatever component receiving that updated information.
        """

        assert task_dict['uid'] == self.uid, 'update called on wrong instance'

        # this method relies on state updates to arrive in order
        current = self.state
        target  = task_dict['state']

        # we never update once the task is FAILED or DONE
        if current in [rps.FAILED, rps.DONE]:
            self._log.debug('task %s is final, ignore update', self.uid)
            return

        # when in CANCELED state, we only allow updates for `DONE` tasks - in
        # that case the cancel command raced the task execution, and the
        # execution actually won, so we don't want to waste that work
        if current == rps.CANCELED and target != rps.DONE:
            self._log.debug('task %s was CANCELED, state not updated', self.uid)
            target = current


        if not reconnect:
            if target not in [rps.FAILED, rps.CANCELED]:
                s_tgt = rps._task_state_value(target)
                s_cur = rps._task_state_value(current)
                if s_tgt - s_cur != 1:
                    self._log.error('%s: invalid state transition %s -> %s',
                                    self.uid, current, target)
                    raise RuntimeError('invalid state transition %s: %s -> %s'
                            % (self.uid, current, target))

        # we update all fields
        # FIXME: well, not all really :/
        # FIXME: setattr is ugly...  we should maintain all state in a dict.
        for key in ['state', 'stdout', 'stderr', 'exit_code', 'return_value',
                    'endpoint_fs', 'resource_sandbox', 'session_sandbox',
                    'pilot', 'pilot_sandbox', 'task_sandbox', 'client_sandbox',
                    'exception', 'exception_detail', 'slots', 'partition',
                    'ofiles']:

            val = task_dict.get(key, None)
            if val is not None:
                setattr(self, "_%s" % key, val)

        metadata = task_dict.get('description', {}).get('metadata')
        if metadata:
            self._descr['metadata'] = metadata

        # if a service is finalized, set info_wait event (only # once)
        if target in rps.FINAL and current not in rps.FINAL:
            if self._descr.mode == TASK_SERVICE:
                # signal failure in case we are still waiting for the service
                self._set_info(None)

        # callbacks are not invoked here, but are bulked in the tmgr


    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a Python dictionary representation of the object."""

        ret = {
            'type':             'task',
            'tmgr':             self.tmgr.uid,
            'uid':              self.uid,
            'name':             self.name,
            'state':            self.state,
            'origin':           self.origin,
            'exit_code':        self.exit_code,
            'stdout':           self.stdout,
            'stderr':           self.stderr,
            'return_value':     self.return_value,
            'exception':        self.exception,
            'exception_detail': self.exception_detail,
            'pilot':            self.pilot,
            'endpoint_fs':      self.endpoint_fs,
            'resource_sandbox': self.resource_sandbox,
            'session_sandbox':  self.session_sandbox,
            'pilot_sandbox':    self.pilot_sandbox,
            'task_sandbox':     self.task_sandbox,
            'client_sandbox':   self.client_sandbox,
            'info':             self.info,
            'slots':            self.slots,
            'partition':        self.partition,
            'description':      self.description,   # this is a deep copy
        }

        return ret


    # --------------------------------------------------------------------------
    #
    @property
    def session(self):
        """radical.pilot.Session: The task's session."""
        return self._session


    # --------------------------------------------------------------------------
    #
    @property
    def tmgr(self):
        """radical.pilot.TaskManager: The task's manager."""
        return self._tmgr


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        """str: The task's unique identifier within a :class:`TaskManager`."""
        return self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        """str: The task's application specified name."""
        return self._descr.name


    # --------------------------------------------------------------------------
    #
    @property
    def mode(self):
        """str: The task mode."""
        return self._descr.mode


    # --------------------------------------------------------------------------
    #
    @property
    def origin(self):
        """str: Indicates where the task was created."""
        return self._origin


    # --------------------------------------------------------------------------
    #
    @property
    def state(self):
        """str: The current state of the task."""
        return self._state


    # --------------------------------------------------------------------------
    #
    @property
    def exit_code(self):
        """int: The exit code of the task, if that is already known, or
        'None' otherwise.
        """
        return self._exit_code


    # --------------------------------------------------------------------------
    #
    @property
    def stdout(self):
        """str: A snapshot of the executable's STDOUT stream.

        If this property is queried before the task has reached
        'DONE' or 'FAILED' state it will return None.

        Warning:
            This can be inefficient.  Output may be incomplete and/or filtered.

        """
        return self._stdout


    # --------------------------------------------------------------------------
    #
    @property
    def stderr(self):
        """str: A snapshot of the executable's STDERR stream.

        If this property is queried before the task has reached
        'DONE' or 'FAILED' state it will return None.

        Warning:
            This can be inefficient.  Output may be incomplete and/or filtered.

        """
        return self._stderr


    # --------------------------------------------------------------------------
    #
    @property
    def output_files(self):
        """list[str]: A list of output file names.

        If this property is queried before the task has reached
        'DONE' or 'FAILED' state it will return None.

        Warning:
            This can be incomplete: the heuristics will not detect files which
            start with `<task_id>.`, for example.  It will also not detect files
            which are not created in the task sandbox.

        """
        return self._ofiles


    # --------------------------------------------------------------------------
    #
    @property
    def return_value(self):
        """Any: The return value for tasks which represent function call (or
        None otherwise).

        If this property is queried before the task has reached
        'DONE' or 'FAILED' state it will always return None.
        """
        return self._return_value


    # --------------------------------------------------------------------------
    #
    @property
    def exception(self):
        """str: A string representation (`__repr__`) of the exception which
        caused the task's `FAILED` state if such one was raised while managing
        or executing the task.

        If this property is queried before the task has reached
        'DONE' or 'FAILED' state it will always return None.
        """
        return self._exception


    # --------------------------------------------------------------------------
    #
    @property
    def exception_detail(self):
        """str: Additional information about the exception which caused this
        task to enter FAILED state.

        If this property is queried before the task has reached
        'DONE' or 'FAILED' state it will always return None.
        """
        return self._exception_detail


    # --------------------------------------------------------------------------
    #
    @property
    def pilot(self):
        """str: The pilot ID of this task, if that is already known, or
        'None' otherwise.
        """
        return self._pilot


    # --------------------------------------------------------------------------
    #
    @property
    def sandbox(self):
        """str: An alias for :attr:`~radical.pilot.Task.task_sandbox`."""
        return self.task_sandbox


    @property
    def task_sandbox(self):
        """radical.utils.Url: The full sandbox URL of this task, if that is already
        known, or 'None' otherwise.
        """
        return self._task_sandbox


    @property
    def endpoint_fs(self):
        """radical.utils.Url: The URL which is internally used to access the
        target resource's root file system.
        """
        return self._endpoint_fs


    @property
    def resource_sandbox(self):
        """radical.utils.Url: The full URL of the path that RP considers the
        resource sandbox, i.e., the sandbox on the target resource's file system
        that is shared by all sessions which access that resource.
        """
        return self._resource_sandbox


    @property
    def session_sandbox(self):
        """radical.utils.Url: The full URL of the path that RP considers the
        session sandbox on the target resource's file system which is shared
        by all pilots which access that resource in the current session.
        """
        return self._session_sandbox


    @property
    def pilot_sandbox(self):
        """radical.utils.Url: The full URL of the path that RP considers the
        pilot sandbox on the target resource's file system which is shared
        by all tasks which are executed by that pilot.
        """
        return self._pilot_sandbox


    @property
    def client_sandbox(self):
        """radical.utils.Url: The full URL of the client sandbox, which is
        usually the same as the current working directory of the Python script
        in which the RP Session is instantiated.

        Note that the URL may not be usable to access
        that sandbox from another machine: RP in general knows nothing about
        available access endpoints on the local host.
        """
        return self._client_sandbox


    # --------------------------------------------------------------------------
    #
    @property
    def description(self):
        """dict: The description the task was started with, as a dictionary."""
        return self._descr


    # --------------------------------------------------------------------------
    #
    @property
    def slots(self):
        '''dict: The slots assigned for the task's execution'''
        if self._slots:
            if not self._slots[0].get('version'):
                for idx,slot in enumerate(self._slots):
                    self._slots[idx] = Slot(self._slots[idx])
        return self._slots


    # --------------------------------------------------------------------------
    #
    @property
    def partition(self):
        '''dict: The pilot partition assigned for the task's execution'''
        return self._partition


    # --------------------------------------------------------------------------
    #
    @property
    def metadata(self):
        """The metadata field of the task's description."""
        return self._descr.metadata


    # --------------------------------------------------------------------------
    #
    @property
    def info(self):
        """The metadata field of the task's description."""
        return self._info

    def _set_info(self, info):
        self._info = info
        self._info_evt.set()

    def wait_info(self, timeout=None):

        self._info_evt.wait(timeout=timeout)

        if self._info is None:
            raise RuntimeError('no service info: %s / %s'
                              % (self.stdout, self.stderr))
        return self.info


    # --------------------------------------------------------------------------
    #
    def register_callback(self, cb, cb_data=None, metric=None):
        """Add a state-change callback.

        Registers a callback function that is triggered every time a
        task's state changes.

        All callback functions need to have the same signature::

            def cb(obj, state) -> None:
                ...

        where ``obj`` is a handle to the object that triggered the callback
        and ``state`` is the new state of that object.  If ``cb_data`` is given,
        then the ``cb`` signature changes to
        ::

            def cb(obj, state, cb_data) -> None:
                ...

        and ``cb_data`` are passed unchanged.
        """

        if not metric:
            metric = rpc.TASK_STATE

        self._tmgr.register_callback(cb, cb_data, metric=metric, uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def wait(self, state=None, timeout=None):
        """Block for state change.

        Returns when the task reaches a specific state or
        when an optional timeout is reached.

        Arguments:
            state (str | list[str], optional): The state(s) that task has to reach
                in order for the call to return.

                By default `wait` waits for the task to reach a **final**
                state, which can be one of the following.

                * :data:`rp.states.DONE`
                * :data:`rp.states.FAILED`
                * :data:`rp.states.CANCELED`
            timeout (float, optional): Optional timeout in seconds before the
                call returns regardless whether the task has reached the desired
                state or not.  The default value **None** never times out.

        """

        if not state:
            states = rps.FINAL
        if not isinstance(state, list):
            states = [state]
        else:
            states = state


        if self.state in rps.FINAL:
            # we will never see another state progression.  Raise an error
            # (unless we waited for this)
            if self.state in states:
                return self.state

            # FIXME: do we want a raise here, really?  This introduces a race,
            #        really, on application level
            # raise RuntimeError("can't wait on a task in final state")
            return self.state

        start_wait = time.time()
        while self.state not in states:

            time.sleep(0.1)

            if timeout and (timeout <= (time.time() - start_wait)):
                break

            if self._tmgr._terminate.is_set():
                break

        return self.state


    # --------------------------------------------------------------------------
    #
    def cancel(self):
        """Cancel the task."""

        self._tmgr.cancel_tasks(self.uid)


# ------------------------------------------------------------------------------
#
class TaskDict(rpu.FastTypedDict):
    """Dictionary encoded Task.

    rp.Task is an API level object and as that is not a useful internal
    representation of an task on the level of RP components and message
    channels.  Instead, a task is there represented as a dictionary.  To
    facilitate a minimum of documentation and type consistency, this class
    defines such task dictionaries as `TypedDict` objects.
    """

    _schema = {
            # where the task got created
            # CLIENT | AGENT
            'origin'      : str,

            # original task description used to create the task
            'description' : TaskDescription,

          # # what scope should be notified for task state changes
          # # CLIENT, AGENT, RAPTOR - can be empty
          # 'notification': [str],

          # # what exact resources were used to execute the task
          # # [[node_uid:core_id:gpu_id, ...]]
          # 'resources'   : [None],
    }

    _defaults = {
            'origin'      : None,
            'description' : None,
          # 'notification': [],
          # 'resources'   : [],
    }


# ------------------------------------------------------------------------------




__copyright__ = "Copyright 2014-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import time
import radical.utils  as ru

from .. import utils as rpu

from .. import Session


# ------------------------------------------------------------------------------
#
class Agent_n(rpu.AgentComponent):

    # This is a sub-agent.  It does not do much apart from starting
    # agent components and watching them, which is all taken care of in the
    # `AgentComponent` base class.

    # --------------------------------------------------------------------------
    #
    def __init__(self, sid, reg_addr, uid):

        reg = ru.zmq.RegistryClient(url=reg_addr)
        cfg = ru.Config(cfg=reg['cfg'])

        # use our own config sans agents/components/bridges as a basis for
        # the sub-agent config.
        a_cfg = cfg['agents'][uid]

        self._uid     = uid
        self._session = Session(uid=sid, cfg=a_cfg,
                                _reg_addr=reg_addr,
                                _role=Session._AGENT_N)

        self._pid     = self._session.cfg.pid
        self._sid     = self._session.cfg.sid
        self._owner   = self._session.cfg.owner
        self._pmgr    = self._session.cfg.pmgr
        self._pwd     = self._session.cfg.pilot_sandbox

        # init the agent component base classes, connects registry
        super().__init__(self._cfg, self._session)



    # --------------------------------------------------------------------------
    def finalize(self):

        self._prof.prof('sub_agent_stop', uid=self._pid, msg=self._uid)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import radical.pilot as rp


verbose  = os.environ.get('RADICAL_PILOT_VERBOSE', 'REPORT')
os.environ['RADICAL_PILOT_VERBOSE'] = verbose

""" DESCRIPTION: Tutorial 2: Chaining Tasks.
For every task A_n a task B_n is started consecutively.
"""

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    RESOURCE_LABEL    = None
    PILOT_CORES       = None
    NUMBER_CHAINS     = None
    TASK_A_EXECUTABLE = None
    TASK_B_EXECUTABLE = None
    QUEUE             = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # Change the user name below if you are using a remote resource
        # and your username on that resource is different from the username
        # on your local machine.
        #


        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        print("Initializing Pilot Manager ...")
        pmgr = rp.PilotManager(session=session)


        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # If you want to run this example on your local machine, you don't have
        # to change anything here.
        #
        # Change the resource below if you want to run on a remote resource.
        # You also might have to set the 'project' to your allocation ID if
        # your remote resource does compute time accounting.
        #
        # A list of preconfigured resources can be found at:
        # https://radicalpilot.readthedocs.io/en/stable/ \
        #        machconf.html#preconfigured-resources
        #
        pdesc = rp.PilotDescription ()
        pdesc.resource = RESOURCE_LABEL
        pdesc.runtime  = 30
        pdesc.cores    = PILOT_CORES
        pdesc.cleanup  = True

        # submit the pilot.
        print("Submitting Pilot to Pilot Manager ...")
        pilot = pmgr.submit_pilots(pdesc)

        # Combine the Pilot, the Tasks and a scheduler via
        # a TaskManager object.
        print("Initializing Task Manager ...")
        tmgr = rp.TaskManager (session=session)


        # Add the created Pilot to the TaskManager.
        print("Registering  Pilot with Task Manager ...")
        tmgr.add_pilots(pilot)

        # submit A tasks to pilot job
        taskdesc_list_A = []
        for i in range(NUMBER_CHAINS):

            # -------- BEGIN USER DEFINED Task A_n DESCRIPTION --------- #
            taskdesc = rp.TaskDescription()
            taskdesc.environment = {"TASK_LIST": "A", "TASK_NO": "%02d" % i}
            taskdesc.executable  = TASK_A_EXECUTABLE
            taskdesc.arguments   = ['"$TASK_LIST Task with id $TASK_NO"']
            taskdesc.cores       = 1
            # -------- END USER DEFINED Task A_n DESCRIPTION --------- #

            taskdesc_list_A.append(taskdesc)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        print("Submit 'A' Tasks to Task Manager ...")
        task_list_A = tmgr.submit_tasks(taskdesc_list_A)

        # Chaining tasks i.e submit a task, when task from A is
        # successfully executed.  A B Task reads the content of the output file of
        # an A Task and writes it into its own output file.
        task_list_B = []

        # We create a copy of task_list_A so that we can remove elements from it,
        # and still reference to the original index.
        task_list_A_copy = task_list_A[:]
        while task_list_A:
            for task_a in task_list_A:
                idx = task_list_A_copy.index(task_a)

                task_a.wait ()
                print("'A' Task '%s' done. Submitting 'B' Task ..." % idx)

                # -------- BEGIN USER DEFINED Task B_n DESCRIPTION --------- #
                taskdesc = rp.TaskDescription()
                taskdesc.environment = {'TASK_LIST': 'B', 'TASK_NO': "%02d" % idx}
                taskdesc.executable  = TASK_B_EXECUTABLE
                taskdesc.arguments   = ['"$TASK_LIST Task with id $TASK_NO"']
                taskdesc.cores       = 1
                # -------- END USER DEFINED Task B_n DESCRIPTION --------- #

                # Submit Task to Pilot Job
                task_b = tmgr.submit_tasks(taskdesc)
                task_list_B.append(task_b)
                task_list_A.remove(task_a)

        print("Waiting for 'B' Tasks to complete ...")
        for task_b in task_list_B :
            task_b.wait ()
            print("'B' Task '%s' finished with output:" % (task_b.uid))
            print(task_b.stdout)

        print("All Tasks completed successfully!")

    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


# ------------------------------------------------------------------------------





# ------------------------------------------------------------------------------
# we *first* import radical.utils, so that the monkeypatching of the logger has
# a chance to kick in before the logging module is pulled by any other 3rd party
# module, and also to monkeypatch `os.fork()` for the `atfork` functionality.
# we also get the version string at this point.
#
import os


# ------------------------------------------------------------------------------
#
# ensure sufficient system limits on OS-X
#
if os.uname()[0] == 'Darwin':
    # on MacOS, we are running out of file descriptors soon.  The code
    # below attempts to increase the limit of open files - but any error
    # is silently ignored, so this is an best-effort, no guarantee.  We
    # leave responsibility for system limits with the user.
    #
    # FIXME: should we do this on all systems, not only Darwin (see
    #        PRTE on Summit)
    try:
        import resource
        _limits    = list(resource.getrlimit(resource.RLIMIT_NOFILE))
        _limits[0] = 512
        resource.setrlimit(resource.RLIMIT_NOFILE, tuple(_limits))

    except:
        pass


# ------------------------------------------------------------------------------
#
from .prof_utils        import *
from .misc              import *
from .session           import *
from .component         import *
from .component_manager import *
from .serializer        import *
from .staging_helper    import *


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2013-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import radical.utils as ru

from .task_description import TaskDescription
from .utils.misc       import FastTypedDict

# ------------------------------------------------------------------------------
# Attribute description keys
#

UID               = 'uid'
RESOURCE          = 'resource'
ACCESS_SCHEMA     = 'access_schema'
QUEUE             = 'queue'
JOB_NAME          = 'job_name'
PROJECT           = 'project'
SANDBOX           = 'sandbox'
RUNTIME           = 'runtime'
APP_COMM          = 'app_comm'         # FIXME: always create an app pubsub?
CLEANUP           = 'cleanup'
EXIT_ON_ERROR     = 'exit_on_error'
SERVICES          = 'services'
ENABLE_EP         = 'enable_ep'
RECONFIG_SRC      = 'reconfig_src'

NODES             = 'nodes'
CORES             = 'cores'
GPUS              = 'gpus'
MEMORY            = 'memory'

INPUT_STAGING     = 'input_staging'
OUTPUT_STAGING    = 'output_staging'

PREPARE_ENV       = 'prepare_env'


# ------------------------------------------------------------------------------
#
class PilotDescription(FastTypedDict):
    """Specify a requested Pilot.

    A PilotDescription object describes the requirements and properties
    of a :class:`radical.pilot.Pilot` and is passed as a parameter to
    :func:`radical.pilot.PilotManager.submit_pilots` to instantiate and run
    a new pilot.

    Note:
        A PilotDescription **MUST** define at least
        :attr:`resource`, :attr:`cores` or :attr:`nodes`,
        and :attr:`runtime`.

    Example::
        pm = radical.pilot.PilotManager(session=s)
        pd = radical.pilot.PilotDescription()
        pd.resource = "local.localhost"
        pd.cores    = 16
        pd.runtime  = 5 # minutes

        pilot = pm.submit_pilots(pd)

    Attributes:
        uid (str, optional): A unique ID for the pilot. A unique ID will be
            assigned by RP if the field is not set.
        job_name (str, optional):  The name of the job / pilot as provided to
            the batch system. If not set then :attr:`uid` will be used instead.
        resource (str): The key of a
            :ref:`platform description </tutorials/configuration.ipynb#Platform-description>`
            entry. If the key exists, the machine-specific
            configuration is loaded from the config file once the `PilotDescription`
            is passed to :meth:`radical.pilot.PilotManager.submit_pilots`. If the
            key doesn't exist, an exception :class:`ValueError` is raised.
        access_schema (str, optional): The key of an access mechanism to use.
            The valid access mechanism is defined in the resource configuration.
            See :doc:`/tutorials/configuration`. The first ``schema`` defined
            in the resource configuration is used by
            default, if no *access_schema* is specified.
        runtime (int, optional): The maximum run time (wall-clock time) in **minutes** of
            the pilot. Default 10.
        sandbox (str, optional): The working ("sandbox") directory of the pilot agent.
            This parameter is optional and if not set, it defaults to
            *radical.pilot.sandbox* in your home or login directory. Default None.

            Warning:
                If you define a pilot on an HPC cluster and you want to set
                `sandbox` manually, make sure that it points to a directory
                on a shared filesystem that can be reached from all
                compute nodes.

        nodes (int, optional): The number of nodes the pilot should allocate on the target
            resource. This parameter could be set instead of `cores` and `gpus`
            (and `memory`). Default 1.

            Note:
                If `nodes` is specified, `gpus` and `cores` must not be specified.

        cores (int, optional): The number of cores the pilot should allocate on
            the target resource. This parameter could be set instead of `nodes`.

            Note:
                For local pilots, you can set a number larger than the physical
                machine limit (corresponding resource configuration should have
                the attribute `"fake_resources"`).

            Note:
                If `cores` is specified, `nodes` must not be specified.

        gpus (int, optional): The number of gpus the pilot should allocate on
            the target resource.

            Note:
                If `gpus` is specified, `nodes` must not be specified.

        memory (int, optional): The total amount of physical memory the
            pilot (and related to it job) requires.
        queue (str, optional): The name of the job queue the pilot should get
            submitted to. If *queue* is set in the resource configuration
            (:attr:`resource`), defining `queue` will override it explicitly.
        project (str, optional): The name of the project / allocation to
            charge for used CPU time. If *project* is set in the resource
            configuration (:attr:`resource`), defining `project` will override it
            explicitly.
        app_comm (list[str], optional): The list of names is interpreted as communication
            channels to start within the pilot agent, for the purpose of
            application communication, i.e., that tasks running on that pilot are
            able to use those channels to communicate amongst each other.

            The names are expected to end in *_queue* or *_pubsub*, indicating the
            type of channel to create.  Once created, tasks will find environment
            variables of the name ``RP_%s_IN`` and ``RP_%s_OUT``, where ``%s``
            is replaced with the given channel name (uppercased), and ``IN/OUT``
            indicate the respective endpoint addresses for the created channels.
        input_staging (list, optional): The list of files to be staged into the
            pilot sandbox.
        output_staging (list, optional): The list of files to be staged from the
            pilot sandbox.
        cleanup (bool, optional): If cleanup is set to True, the pilot
            will delete its entire sandbox upon termination. This includes individual
            Task sandboxes and all generated output data. Only log files will
            remain in the sandbox directory. Default False.
        exit_on_error (bool, optional): Flag to trigger app termination in case
            of the pilot failure. Default True.
        services (list[TaskDescription], optional): A list of
            commands which get started on a separate service compute node right after
            bootstrapping, and before any RP task is launched.  That service compute
            node will not be used for any other tasks.
        enable_ep (bool, optional): enable a ZMQ submission endpoint on the
            pilot. Default `False`.
        prepare_env (dict, optional): A dictionary of `{env_name: env_spec}` as
            documented for the pilot's `prepare_env(env_name, env_spec)` method.
            The given specifications will be enacted during pilot startup and
            can be used for service tasks.
        reconfig_src (string, optional): name of a data file to be used by the
            agent's `reconfig` scheduler.

    """

    _schema = {
        UID             : str        ,
        RESOURCE        : str        ,
        ACCESS_SCHEMA   : str        ,
        RUNTIME         : int        ,
        APP_COMM        : [str]      ,
        SANDBOX         : str        ,
        CORES           : int        ,
        GPUS            : int        ,
        NODES           : int        ,
        MEMORY          : int        ,
        QUEUE           : str        ,
        JOB_NAME        : str        ,
        PROJECT         : str        ,
        CLEANUP         : bool       ,
        EXIT_ON_ERROR   : bool       ,
        INPUT_STAGING   : [str]      ,
        OUTPUT_STAGING  : [str]      ,
        PREPARE_ENV     : {str: None},
        SERVICES        : [TaskDescription],
        ENABLE_EP       : bool       ,
        RECONFIG_SRC    : str        ,
    }

    _defaults = {
        UID             : None       ,
        RESOURCE        : None       ,
        ACCESS_SCHEMA   : None       ,
        RUNTIME         : 10         ,
        APP_COMM        : []         ,
        SANDBOX         : None       ,
        CORES           : 0          ,
        GPUS            : 0          ,
        NODES           : 0          ,
        MEMORY          : 0          ,
        QUEUE           : None       ,
        JOB_NAME        : None       ,
        PROJECT         : None       ,
        CLEANUP         : False      ,
        EXIT_ON_ERROR   : True       ,
        INPUT_STAGING   : []         ,
        OUTPUT_STAGING  : []         ,
        PREPARE_ENV     : {}         ,
        SERVICES        : []         ,
        ENABLE_EP       : False      ,
        RECONFIG_SRC    : None       ,
    }


    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None):

        super().__init__(from_dict=from_dict)


    # --------------------------------------------------------------------------
    #
    def _verify(self):

        if not self.get('resource'):
            raise ValueError("Pilot description needs 'resource'")

        if self.get('nodes'):

            if self.get('cores'):
                raise ValueError("Pilot description needs 'cores' *or* 'nodes'")

            if self.get('gpus'):
                raise ValueError("Pilot description needs 'cores' *or* 'nodes'")


# ------------------------------------------------------------------------------




__copyright__ = 'Copyright 2016-2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

from typing import List, Tuple

import radical.utils as ru

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class PBSPro(ResourceManager):

    # --------------------------------------------------------------------------
    #
    @staticmethod
    def batch_started():

        return bool(os.getenv('PBS_JOBID'))

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        nodes = None

        try:
            vnodes, rm_info.cores_per_node = self._parse_pbspro_vnodes()
            nodes = [(node, rm_info.cores_per_node) for node in vnodes]

        except (IndexError, ValueError):
            self._log.debug_2('exec_vnodes not detected')

        except RuntimeError as e:
            err_message = str(e)
            if not err_message.startswith('qstat failed'):
                raise
            self._log.debug_1(err_message)

        if not nodes:

            if not rm_info.cores_per_node or 'PBS_NODEFILE' not in os.environ:
                raise RuntimeError('resource configuration unknown, either '
                                   'cores_per_node or $PBS_NODEFILE not set')

            nodes = self._parse_nodefile(os.environ['PBS_NODEFILE'],
                                         cpn=rm_info.cores_per_node,
                                         smt=rm_info.threads_per_core)

        rm_info.node_list = self._get_node_list(nodes, rm_info)

        return rm_info

    # --------------------------------------------------------------------------
    #
    def _parse_pbspro_vnodes(self) -> Tuple[List[str], int]:

        # PBS Job ID
        jobid = os.environ.get('PBS_JOBID')
        if not jobid:
            raise RuntimeError('$PBS_JOBID not set')

        # Get the output of qstat -f for this job
        output, error, ret = ru.sh_callout(['qstat', '-f', jobid])
        if ret:
            raise RuntimeError('qstat failed: %s' % error)

        # Get the (multiline) 'exec_vnode' entry
        vnodes_str = ''
        for line in output.splitlines():
            line = ru.as_string(line)
            # Detect start of entry
            if 'exec_vnode = ' in line:
                vnodes_str += line.strip()
            elif vnodes_str:
                # Find continuing lines
                if ' = ' in line:
                    break
                vnodes_str += line.strip()

        # Get the RHS of the entry
        rhs = vnodes_str.split('=', 1)[1].strip()
        self._log.debug_1('exec_vnodes: %s', rhs)

        nodes_list = []
        # Break up the individual node partitions into vnode slices
        while True:
            idx = rhs.find(')+(')
            node_str = rhs[1:idx]
            nodes_list.append(node_str)
            rhs = rhs[idx + 2:]
            if idx < 0:
                break

        vnodes_set = set()
        ncpus_set  = set()
        # Split out the slices into vnode name and cpu count
        for node_str in nodes_list:
            slices = node_str.split('+')
            for _slice in slices:
                vnode, ncpus = _slice.split(':')
                vnodes_set.add(vnode)
                ncpus_set.add(int(ncpus.split('=')[1]))

        self._log.debug_1('vnodes: %s', vnodes_set)
        self._log.debug_1('ncpus: %s',  ncpus_set)

        if len(ncpus_set) > 1:
            raise RuntimeError('detected vnodes of different sizes')

        return sorted(vnodes_set), ncpus_set.pop()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

import radical.utils as ru

from ...   import states             as rps
from ...   import constants          as rpc
from ...   import utils              as rpu

from ...staging_directives import expand_staging_directives

from .base import TMGRStagingOutputComponent



# ------------------------------------------------------------------------------
#
class Default(TMGRStagingOutputComponent):
    """
    This component performs all tmgr side output staging directives for compute
    tasks.  It gets tasks from the tmgr_staging_output_queue, in
    TMGR_STAGING_OUTPUT_PENDING state, will advance them to TMGR_STAGING_OUTPUT
    state while performing the staging, and then moves then to the respective
    final state.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        TMGRStagingOutputComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._stager = rpu.StagingHelper(self._log)

        self.register_input(rps.TMGR_STAGING_OUTPUT_PENDING,
                            rpc.PROXY_TASK_QUEUE,
                            qname=self._session.uid,
                            cb=self.work)

        # we don't need an output queue -- tasks will be final


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        if not isinstance(tasks, list):
            tasks = [tasks]

        self.advance(tasks, rps.TMGR_STAGING_OUTPUT, publish=True, push=False)

        # we first filter out any tasks which don't need any output staging, and
        # advance them again as a bulk.  We work over the others one by one, and
        # advance them individually, to avoid stalling from slow staging ops.

        no_staging_tasks = list()
        staging_tasks    = list()

        for task in tasks:

            # we only handle staging for tasks which end up in `DONE` state
            target_state = task.get('target_state')
            if target_state and target_state != rps.DONE:
                self._log.debug('skip staging for %s', task['uid'])
                no_staging_tasks.append(task)
                continue

            # check if we have any staging directives to be enacted in this
            # component
            actionables = list()
            for sd in task['description'].get('output_staging', []):

                if sd['action'] == rpc.TRANSFER:
                    actionables.append(sd)

            if actionables:
                staging_tasks.append([task, actionables])
            else:
                no_staging_tasks.append(task)


        if no_staging_tasks:

            # nothing to stage -- transition into final state.
            for task in no_staging_tasks:
                task['state'] = task['target_state']
            self.advance(no_staging_tasks, publish=True, push=True)

        for task,actionables in staging_tasks:
            try:
                self._handle_task(task, actionables)
                self.advance(task, publish=True, push=True)
            except:
                self._log.exception("staging error")
                self.advance(task, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task, actionables):

        uid = task['uid']

        src_context = {'pwd'      : task['task_sandbox'],       # !
                       'client'   : task['client_sandbox'],
                       'task'     : task['task_sandbox'],
                       'pilot'    : task['pilot_sandbox'],
                       'session'  : task['session_sandbox'],
                       'resource' : task['resource_sandbox'],
                       'endpoint' : task['endpoint_fs']}
        tgt_context = {'pwd'      : task['client_sandbox'],     # !
                       'client'   : task['client_sandbox'],
                       'task'     : task['task_sandbox'],
                       'pilot'    : task['pilot_sandbox'],
                       'session'  : task['session_sandbox'],
                       'resource' : task['resource_sandbox'],
                       'endpoint' : task['endpoint_fs']}

        # url used for cache (sandbox url w/o path)
        tmp      = ru.Url(task["task_sandbox"])
        tmp.path = '/'
        key      = str(tmp)

        actionables = expand_staging_directives(actionables,
                                            src_context, tgt_context, self._log)

        # Loop over all transfer directives and execute them.
        for sd in actionables:
            self._prof.prof('staging_in_start', uid=uid, msg=sd['uid'])
            self._stager.handle_staging_directive(sd)
            self._prof.prof('staging_in_stop', uid=uid, msg=sd['uid'])

        # all staging is done -- at this point the task is final
        task['state'] = task['target_state']
        self.advance(task, publish=True, push=True)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import radical.pilot as rp


def state_cb(task, state):
    if state in rp.FINAL:
        print('%s [%d]: %s' % (task.uid, task.description.priority, task.state))


if __name__ == '__main__':

    session = rp.Session()
    try:

        pmgr  = rp.PilotManager(session=session)
        tmgr  = rp.TaskManager(session=session)
        pdesc = rp.PilotDescription({'resource': 'local.localhost',
                                     'runtime' : 1024 * 1024,
                                     'nodes'   : 1})
        pilot = pmgr.submit_pilots(pdesc)

        tmgr.add_pilots(pilot)
        tmgr.register_callback(state_cb)

        n   = 200
        tds = list()
        for i in range(n):
            td = rp.TaskDescription({'executable': 'true',
                                     'priority'  : i % 3})
            tds.append(td)

        tasks = tmgr.submit_tasks(tds)

        tmgr.wait_tasks()

    finally:
        session.close(download=True)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2024, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import time

import threading     as mt
import subprocess    as sp

import radical.utils as ru

from ...  import states    as rps
from ...  import constants as rpc

from .popen import Popen


# ------------------------------------------------------------------------------
#
class Dragon(Popen):

    # --------------------------------------------------------------------------
    #
    def initialize(self):
        '''
        This method is called by the base class during component
        initialization.

        It will start the dragon executor. The executor will communicate the endpoint
        URLs back to us. We will then use those endpoints to communicate with the dragon
        process.
        '''

        super().initialize()

        self._pid = self.session.cfg.pid

        # run dragon execution helper in a separate process which actually uses
        # dragon's multiprocessing implementation
        self._dragon = sp.Popen(
                args   = ['dragon', 'radical-pilot-dragon-executor.py',
                          self.session._cfg.path],
                stdout = sp.PIPE,
                stderr = sp.STDOUT)

        # run this loop for 10 seconds to get the endpoints
        start = time.time()
        while True:

            if self._dragon.poll() is not None:
                self._log.error('%s', str(self._dragon.communicate()))
                raise RuntimeError('dragon process died')

            line = ru.as_string(self._dragon.stdout.readline().strip())
            self._log.debug('line: [%s]', line)

            if line.startswith('ZMQ_ENDPOINTS '):
                _, url_out, url_in = line.split()
                break

            if time.time() - start > 3:
                self._log.debug('%s', str(self._dragon.communicate()))
                raise RuntimeError('failed to get dragon endpoint')

        self._log.debug('dragon eps: %s - %s', url_in, url_out)

        self._pipe_out = ru.zmq.Pipe(ru.zmq.MODE_PUSH, url_out)

        # run watcher thread
        self._watcher = mt.Thread(target=self._dragon_watch, args=[url_in])
        self._watcher.daemon = True
        self._watcher.start()


    # --------------------------------------------------------------------------
    #
    def _launch_task(self, task):
        '''
        This method is called by the base class to actually
        launch the task.
        '''

        # send task to dragon
        self._log.debug('launch task %s', task['uid'])
        self._pipe_out.put({'cmd': 'run', 'task': task})

        # handle task timeout if needed
        self.handle_timeout(task)


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, task):
        '''
        This method is called by the base class to actually
        cancel the task.
        '''

        # send cancel request to dragon
        self._pipe_out.put({'cmd': 'cancel', 'uid': uid})


    # --------------------------------------------------------------------------
    #
    def _dragon_watch(self, url_in):
        '''
        This method watches the dragon endpoint for task completion messages.

        :param url_in:  the dragon endpoint to watch
        '''

        pipe_in  = ru.zmq.Pipe(ru.zmq.MODE_PULL, url_in)

        try:
            while not self._term.is_set():

                msg = pipe_in.get_nowait(0.1)

                if not msg:
                    continue

                cmd = msg.get('cmd')

                if cmd != 'done':
                    raise ValueError('unsupported command %s' % cmd)

                task = msg['task']
                tid  = task['uid']

                self._prof.prof('unschedule_start', uid=tid)

                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, [task])

                self.advance([task], rps.AGENT_STAGING_OUTPUT_PENDING,
                                     publish=True, push=True)

        except Exception as e:
            self._log.exception('Error in ExecWorker watch loop (%s)' % e)
            self.stop()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import copy

import radical.utils as ru

from .continuous import Continuous

from ... import states    as rps
from ... import constants as rpc


# ------------------------------------------------------------------------------
#
# This is a simple extension of the Continuous scheduler which evaluates the
# `order` tag of arriving tasks, which is expected to have the form
#
#   order : {'ns'   : <string>,
#            'order': <int>,
#            'size' : <int>}
#
# where 'ns' is a namespace, 'order' is an integer defining the order of bag of
# tasks in that namespace, and 'size' is the number of tasks in that bag.  The
# semantics of the scheduler is that, for any given namespace, a BoT with order
# 'n' will only be executed after 'size' tasks of the BoT with order 'n-1' have
# been executed.  The first BoT is expected to have order '0'.
#
# The dominant use case for this scheduler is the execution of pipeline stages,
# where one stage needs to be completed before tasks from the next stage can be
# considered for scheduling.
#
# FIXME: - failed tasks cannot yet be recognized
#
class ContinuousOrdered(Continuous):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        Continuous.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        Continuous._configure(self)

        # This scheduler will wait for state updates, and will consider a task
        # `done` once it reaches a trigger state.  When a state update is found
        # which shows that the tasks reached that state, it is marked as 'done'
        # in the respective order of its namespace.
        #
        self._trigger_state = rps.TMGR_STAGING_OUTPUT_PENDING
        self.register_subscriber(rpc.STATE_PUBSUB, self._state_cb)

        # a namespace entry will look like this:
        #
        #   { 'current' : 0,   # BoT currently operated on - starts with '0'
        #     0 :              # sequential BoT numbering  `n`
        #     {'size': 128,    # number of tasks to expect   `max`
        #      'uids': [...]}, # ids    of tasks to be scheduled
        #      'done': [...]}, # ids    of tasks in trigger state
        #     },
        #     ...
        #   }
        #
        # prepare an initial entry for each ns which ensures that BOT #0 is
        # runnable once it arrives.

        self._lock       = ru.RLock()   # lock on the ns
        self._tasks      = dict()       # task registry (we use uids otherwise)
        self._unordered  = list()       # IDs of tasks which are not ordered
        self._ns         = dict()       # nothing has run, yet

        self._ns_init    = {'current' : 0}
        self._order_init = {'size'    : 0,
                            'uids'    : list(),
                            'done'    : list()}


    # --------------------------------------------------------------------------
    # overload the main method from the base class
    def _schedule_task(self, task):

        self.advance(task, rps.AGENT_SCHEDULING, publish=True, push=False)

        with self._lock:

            # cache ID int to avoid repeated parsing
            uid       = task['uid']
            descr     = task['description']
            order_tag = descr.get('tags', {}).get('order')

            # tasks w/o order info are handled as usual, and we don't keep
            # any infos around
            if not order_tag:
              # self._log.debug('no tags for %s', uid)
                self._unordered.append(task)
                return

            # this uniit wants to be ordered - keep it in our registry
            assert uid not in self._tasks, 'duplicated task %s' % uid
            self._tasks[uid] = task

            ns    = order_tag['ns']
            order = order_tag['order']
            size  = order_tag['size']

          # self._log.debug('tags %s: %s : %d : %d', uid, ns, order, size)
            # initiate ns if needed
            if ns not in self._ns:
                self._ns[ns] = copy.deepcopy(self._ns_init)

            # initiate order if needed
            if order not in self._ns[ns]:
                self._ns[ns][order] = copy.deepcopy(self._order_init)
                self._ns[ns][order]['size'] = size

            # ensure that order information are consistent
            assert size == self._ns[ns][order]['size'], \
                   'inconsistent order size'

            # add task to order
            self._ns[ns][order]['uids'].append(uid)

        # try to schedule known tasks
        self._try_schedule()

        return


    # --------------------------------------------------------------------------
    def _try_schedule(self):
        '''
        Schedule all tasks in self._unordered.  Then for all name spaces,
        check if their `current` order has tasks to schedule.  If not and
        we see `size` tasks are `done`, consider the order completed and go
        to the next one.  Break once we find a BoT which is not completely
        schedulable, either because we did not yet get all its tasks, or
        because we run out of resources to place those tasks.
        '''

      # self._log.debug('try schedule')
        scheduled = list()  # list of scheduled tasks

        # FIXME: this lock is very aggressive, it should not be held over
        #        the scheduling algorithm's activity.
        # first schedule unordered tasks (
        with self._lock:

            keep = list()
            for task in self._unordered:

                # attempt to schedule this task (use continuous algorithm)
                if Continuous._try_allocation(self, task):

                    # success - keep it and try the next one
                    scheduled.append(task)

                else:
                    # failure - keep task around
                    keep.append(task)

            # keep only unscheduleed tasks
            self._unordered = keep


        # FIXME: this lock is very aggressive, it should not be held over
        #        the scheduling algorithm's activity.
        with self._lock:

            # now check all namespaces for eligibility
            for ns in self._ns:

                # seek next eligible order - start with current one
                current = self._ns[ns]['current']
                while True:

                    # any information available for the current order?
                    if current not in self._ns[ns]:
                        break

                    order = self._ns[ns][current]

                    # is this order complete?  If so, advance and go to next
                    if order['size'] == len(order['done']):
                      # self._log.debug('order %s [%s] completed', ns, current)
                        current += 1
                        order['current'] = current
                        continue

                    # otherwise we found the order we want to handle
                    # try to schedule the tasks in this order
                    keep = list()
                    for uid in order['uids']:

                        task = self._tasks[uid]

                        # attempt to schedule this task with the continuous alg
                        if Continuous._try_allocation(self, task):

                            # success - keep it and try the next one
                            scheduled.append(task)

                        else:
                            # failure - keep task around
                            keep.append(uid)

                    # only keep unscheduled tasks around
                    order['uids'] = keep

                    # we always break after scheduling tasks from an order.
                    # Either we are out of resources, then we have to try again
                    # later,  or we are out of tasks, and could switch to the
                    # next order -- but that would require state updates from
                    # the tasks just submitted, so we break anyway
                    break


        # advance all scheduled tasks and push them out
        if scheduled:
            self.advance(scheduled, rps.AGENT_EXECUTING_PENDING,
                         publish=True, push=True)

      # self._log.debug('dump')
      # self._log.debug(pprint.pformat(self._ns))


    # --------------------------------------------------------------------------
    #
    def schedule_cb(self, topic, msg):
        '''
        This cb gets triggered after some tasks got unscheduled, ie. their
        resources have been freed.  We attempt a new round of scheduling at that
        point.
        '''
        self._try_schedule()

        # keep the cb registered
        return True


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, topic, msg):
        '''
        Get state updates, and trigger a new schedule attempt if any task
        previously scheduled by us reaches the trigger state.
        '''

        cmd    = msg.get('cmd')
        things = msg.get('arg')

        if cmd not in ['update']:
            self._log.info('ignore cmd %s', cmd)
            return True

        if not isinstance(things, list):
            things = [things]

        trigger = False
        for thing in things:

            uid   = thing['uid']
            ttype = thing['type']
            state = thing['state']

            if ttype != 'task':
                continue

            # FIXME: this could be optimized by also checking for later states
            if not state or state != self._trigger_state:
                continue

            with self._lock:
                if uid not in self._tasks:
                    # unknown task
                    continue

            # get and parse order tag.  We don't need to repeat checks
            order_tag = thing['description']['tags']['order']
            ns    = order_tag['ns']
            order = order_tag['order']
          # size  = order_tag['size']

            with self._lock:
                self._ns[ns][order]['done'].append(uid)

            trigger = True


        # did any tasks trigger a reschedule?
        if trigger:
            self._try_schedule()

        return True


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import time

import radical.pilot as rp


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # demonstrate task cancelation during execution: we start some tasks, then
    # wait for some to finish, and cancel some more, both tasks currently
    # executing and tasks still pending for execution

    session = rp.Session()

    try:
        pmgr    = rp.PilotManager(session=session)
        pd_init = {'resource': 'local.localhost',
                   'runtime' : 60,
                   'cores'   :  8
                  }
        pdesc = rp.PilotDescription(pd_init)
        pilot = pmgr.submit_pilots(pdesc)
        tmgr  = rp.TaskManager(session=session)
        tmgr.add_pilots(pilot)

        # wait until pilot is active to have somewhat sensible timings below
        pilot.wait(rp.PMGR_ACTIVE)
        print('pilot state: %s' % pilot.state)

        # submit 32 tasks, each running for 10 seconds.  We have 8 cores, so the
        # tasks will run in 4 batches of 8 at a time.
        t_start = time.time()
        tds     = list()
        for _ in range(32):

            td = rp.TaskDescription()
            td.executable = 'radical-pilot-hello.sh'
            td.arguments  = [10]
            tds.append(td)

        tasks = tmgr.submit_tasks(tds)

        # wait until 8 tasks are done (16 cores: should take about 10 seconds
        print('%d total  - wait for 8 tasks' % len(tasks))
        while True:
            time.sleep(1)
            states = [t.state for t in tasks]
            count  = len([s for s in states if s in rp.FINAL])
            if count >= 8:
                break
        print('%d DONE out of %d total' % (len(states), count))

        t_gen_1 = time.time()

        # cancel 12 tasks which are not in the set of DONE tasks above.  That
        # should request cancellation for a mix of currently running and still
        # pending tasks (8 are DONE, 8 are running right now, 16 are pending.
        done_uids   = [t.uid for t in tasks if t.state == rp.DONE]
        cancel_uids = [t.uid for t in tasks if t.uid not in done_uids]
        cancel_uids = cancel_uids[:12]
        n_canceled  = len(cancel_uids)

        tmgr.cancel_tasks(uids=cancel_uids)

        t_cancel = time.time()

        # 12 tasks should be left which fit on the available cores in two
        # batches (8 + 4), and we should wait for about 20 seconds give or take
        tmgr.wait_tasks()
        t_done = time.time()

        # print some statistocs
        states = [t.state for t in tasks]
        print('canceled %d tasks' % n_canceled)
        print('states:')
        for s in [rp.DONE, rp.FAILED, rp.CANCELED]:
            print ('  %-10s: %3d' % (s, states.count(s)))

        print('\ntimes:')
        print('  gen 1 : %10.1f sec (exp: 10 sec)' % (t_gen_1  - t_start))
        print('  cancel: %10.1f sec (exp:  0 sec)' % (t_cancel - t_gen_1))
        print('  gen x : %10.1f sec (exp: 20 sec)' % (t_done   - t_cancel))

        print()
        for task in tasks:
            print('task %s: %s' % (task.uid, task.state))

    finally:
        session.close()


# ------------------------------------------------------------------------------





import os
import time


from collections import defaultdict
from typing      import List

import threading         as mt

import radical.utils     as ru

from .. import utils     as rpu
from .. import states    as rps
from .. import constants as rpc

from .. import Session, Task, TaskDescription, TASK_EXECUTABLE

from ..task_description import RAPTOR_WORKER


# ------------------------------------------------------------------------------
#
class Master(rpu.AgentComponent):
    '''
    Raptor Master class

    The `rp.raptor.Master` instantiates and orchestrates a set of workers which
    are used to rapidly and efficiently execute function tasks.  As such the
    raptor master acts as an RP executor: it hooks into the RP agent
    communication channels to receive tasks from the RP agent scheduler in order
    to execute them.  Once completed tasks are pushed toward the agent output
    staging component and will then continue their life cycle as all other
    tasks.
    '''

    # flags for worker readiness.  These flags have somewhat different meaning
    # than the worker's task state: the worker reaching `AGENT_EXECUTING` is
    # necessary, but the worker also needs to perform some setup steps and needs
    # to hook into the agent's communication channels - only then is the worker
    # considered `ACTIVE` and ready to receive tasks.
    NEW    = 'NEW'
    ACTIVE = 'ACTIVE'
    DONE   = 'DONE'


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg: ru.Config = None):
        '''
        This raptor master is expected to be hosted in a main thread of a RP
        task instance.  As such the normal `RP_*` environment variables are
        expected to be available.

        This c'tor will create communication channels which are later used by
        workers to communicate with this master instance.

        Args:
            cfg: session config.  fallback: agent config
        '''

        self._uid        = os.environ['RP_TASK_ID']
        self._pid        = os.environ['RP_PILOT_ID']
        self._sid        = os.environ['RP_SESSION_ID']
        self._name       = os.environ['RP_TASK_NAME']
        self._sbox       = os.environ['RP_TASK_SANDBOX']
        self._psbox      = os.environ['RP_PILOT_SANDBOX']
        self._ssbox      = os.environ['RP_SESSION_SANDBOX']
        self._rsbox      = os.environ['RP_RESOURCE_SANDBOX']
        self._reg_addr   = os.environ['RP_REGISTRY_ADDRESS']

        self._workers    = dict()      # wid: worker
        self._tasks      = dict()      # bookkeeping of submitted requests
        self._exec_tasks = list()      # keep track of executable tasks
        self._term       = mt.Event()  # termination signal
        self._thread     = None        # run loop

        self._session    = Session(uid=self._sid, _reg_addr=self._reg_addr,
                                   _role=Session._DEFAULT)

        ccfg = ru.Config(from_dict={'uid'     : self._uid,
                                    'sid'     : self._sid,
                                    'owner'   : self._pid,
                                    'reg_addr': self._reg_addr})

        super().__init__(ccfg, self._session)

        # get hb configs (RegistryClient instance is initiated in Session)
        self._hb_freq = self._session.rcfg.raptor.hb_frequency
        self._hb_tout = self._session.rcfg.raptor.hb_timeout

        self._log.debug('hb freq: %s', self._hb_freq)
        self._log.debug('hb tout: %s', self._hb_tout)

        # we never run `self.start()` which is ok - but it means we miss out on
        # some of the component initialization.  Call it manually thus
        self._initialize()

        # register termination handler
        self.register_rpc_handler('raptor_rpc', self._raptor_rpc,
                                  rpc_addr=self.uid)

        # send new worker tasks and agent input staging / agent scheduler
        self.register_output(rps.AGENT_STAGING_INPUT_PENDING,
                             rpc.AGENT_STAGING_INPUT_QUEUE)

        # set up zmq queues between the agent scheduler and this master so that
        # we can receive new requests from RP tasks
        qname = '%s.input_queue' % self._uid
        input_cfg = ru.Config(cfg={'channel'   : qname,
                                   'type'      : 'queue',
                                   'uid'       : '%s_input' % self._uid,
                                   'path'      : self._sbox,
                                   'stall_hwm' : 0,
                                   'bulk_size' : 1})

        # FIXME: how to pass cfg?
        self._input_queue = ru.zmq.Queue(qname, cfg=input_cfg)
        self._input_queue.start()

        # send completed request tasks to agent output staging / tmgr
        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        # set up zmq queues between this master and all workers for request
        # distribution and result collection
        req_cfg = ru.Config(cfg={'channel'   : 'raptor_tasks',
                                 'type'      : 'queue',
                                 'uid'       : self._uid + '.req',
                                 'path'      : self._sbox,
                                 'stall_hwm' : 0,
                                 'bulk_size' : 1})

        res_cfg = ru.Config(cfg={'channel'   : 'raptor_results',
                                 'type'      : 'queue',
                                 'uid'       : self._uid + '.res',
                                 'path'      : self._sbox,
                                 'stall_hwm' : 0,
                                 'bulk_size' : 1})

        self._req_queue = ru.zmq.Queue('raptor_tasks',   cfg=req_cfg)
        self._res_queue = ru.zmq.Queue('raptor_results', cfg=res_cfg)

        self._req_queue.start()
        self._res_queue.start()

        self._req_addr_put = str(self._req_queue.addr_put)
        self._req_addr_get = str(self._req_queue.addr_get)

        self._res_addr_put = str(self._res_queue.addr_put)
        self._res_addr_get = str(self._res_queue.addr_get)

        # this master will put requests onto the request queue, and will get
        # responses from the response queue.  Note that the responses will be
        # delivered via an async callback (`self._result_cb`).
        self._req_put = ru.zmq.Putter('raptor_tasks',    self._req_addr_put)
        self._res_get = ru.zmq.Getter('raptor_results',  self._res_addr_get,
                                                         cb=self._result_cb)

        # also create a ZMQ server endpoint for the workers to
        # send task execution requests back to the master
        self._task_service = ru.zmq.Server()
        self._task_service.register_request('run_task', self._run_task)
        self._task_service.start()
        self._task_service_data = dict()   # task.uid : [mt.Event, task]

        # for the workers it is the opposite: they will get requests from the
        # request queue, and will send responses to the response queue.
        self._info = {'req_addr_get': self._req_addr_get,
                      'res_addr_put': self._res_addr_put,
                      'task_service': self._task_service.addr}

        # make sure the channels are up before allowing to submit requests
        time.sleep(1)

        # begin to receive tasks in that queue
        ru.zmq.Getter(qname, self._input_queue.addr_get, cb=self._request_cb)

        # everything is set up - we can serve messages on the pubsubs also
        self.register_subscriber(rpc.STATE_PUBSUB, self._state_cb)

        # and register that input queue with the scheduler
        self._log.debug('registered raptor queue: %s / %s', self._uid, qname)
        self.publish(rpc.CONTROL_PUBSUB,
                      {'cmd': 'register_raptor_queue',
                       'arg': {'name' : self._uid,
                               'queue': qname,
                               'addr' : str(self._input_queue.addr_put)}})

        # all comm channels are set up - begin to work
        self._log.debug('startup complete')


    # --------------------------------------------------------------------------
    #
    @property
    def workers(self):
        '''
        task dictionaries representing all currently registered workers
        '''
        return self._workers


    # --------------------------------------------------------------------------
    #
    def _raptor_rpc(self, *args, **kwargs):

        self._log.debug('r rpc (%s, %s)', args, kwargs)

        raptor_cmd = kwargs.get('raptor_cmd')

        if raptor_cmd == 'stop':
            self.stop()


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):
        '''
        listen for `worker_register`, `worker_unregister`,
        `worker_rank_heartbeat` and `rpc_req` messages.
        '''

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        if cmd == 'worker_register':

            uid   = arg['uid']
            rid   = arg['raptor_id']
            ranks = arg['ranks']

            if rid != self._uid:
                return

            now = time.time()
            if uid not in self._workers:
                self._workers[uid] = {
                        'uid'        : uid,
                        'status'     : self.NEW,
                        'heartbeats' : {r: now for r in range(ranks)}
                }

            self._workers[uid]['status'] = self.ACTIVE

            # return a message to the worker to inform about the master service
            # endpoints
            info = {'req_addr_get': self._req_addr_get,
                    'res_addr_put': self._res_addr_put,
                    'ts_addr'     : self._task_service.addr}
            self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'worker_registered',
                                              'arg': {'uid' : uid,
                                                      'info': info}})

        elif cmd == 'worker_rank_heartbeat':

            uid  = arg['uid']
            rank = arg['rank']

            self._log.debug('recv rank heartbeat %s:%s', uid, rank)

            if uid not in self._workers:
                return

            self._workers[uid]['heartbeats'][rank] = time.time()


        elif cmd == 'worker_unregister':

            uid = arg['uid']
            self._log.debug('unregister %s', uid)

            if uid not in self._workers:
                return

            self._workers[uid]['status'] = self.DONE



    # --------------------------------------------------------------------------
    #
    def _state_cb(self, topic, msg):
        '''
        listen for state updates for tasks executed by raptor workers, but also
        check for state updates originating directly from our workers.
        '''

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        # state update for tasks created by raptor
        if cmd == 'raptor_state_update':

            # raptor workers can get caught in the `raptor_state_update` hook
            # when they are submitted by the master. Filter them out - we don't
            # want result callbacks for workers.
            tasks = [task for task in arg
                          if  task['description']['mode'] != RAPTOR_WORKER]

            for task in tasks:
                self._log.debug('raptor state update: %s : %s',
                                task['uid'], task['state'])

            self._result_cb(tasks)


        # general task state updates -- check if our workers are affected
        elif cmd == 'update':

            for thing in ru.as_list(arg):

                uid   = thing['uid']
                state = thing['state']

                if uid in self._workers:

                    if state == rps.AGENT_STAGING_OUTPUT:
                        self._workers[uid]['status'] = self.DONE
                        self._log.info('worker %s final: %s', uid, state)

                    self.worker_state_cb(self._workers, state)

        return True


    # --------------------------------------------------------------------------
    #
    def worker_state_cb(self, worker_dict, state):
        '''
        This callback can be overloaded - it will be invoked whenever the master
        receives a state update for a worker it is connected to.

        args:
            worker_dict (Dict[str, Any]): a task dictionary representing the
                worker whose state was updated
            state (str): new state of the worker
        '''

        pass


    # --------------------------------------------------------------------------
    #
    def submit_workers(self, descriptions: List[TaskDescription]
                      ) -> List[str]:
        '''
        Submit a raptor workers per given `descriptions` element and pass the
        queue raptor info as configuration file.  Do *not* wait for the workers
        to come up - they are expected to register via the control channel.

        The task `descriptions` specifically support the following keys:

          - raptor_class: str, type name of worker class to execute
          - raptor_file : str, optional
              Module file from which *raptor_class* may be imported,
              if a custom RP worker class is used

        Note that only one worker rank (presumably rank 0) should register with
        the master - the workers are expected to synchronize their ranks as
        needed.

        Args:
            descriptions (List[TaskDescription]): a list of worker descriptions

        Returns:
            List[str]: list of uids for submitted worker tasks
        '''

        tasks = list()
        for td in descriptions:

            if td.mode != RAPTOR_WORKER:
                raise ValueError('unexpected task mode [%s]' % td.mode)

            # sharing GPUs among multiple ranks not yet supported
            if td.gpus_per_rank and not td.gpus_per_rank.is_integer():
                raise RuntimeError('GPU sharing for workers is not supported')

            raptor_file  = td.get('raptor_file')  or  ''
            raptor_class = td.get('raptor_class') or  'DefaultWorker'

            td.raptor_id = self.uid

            if not td.get('uid'):
                td.uid = '%s.%s' % (self.uid, ru.generate_id('worker',
                                                             ns=self.uid))
            if not td.get('executable'):
                td.executable = 'radical-pilot-raptor-worker'

            if not td.get('sandbox'):
                td.sandbox = self._sbox

            # this master is obviously running in a suitable python3 env,
            # so we expect that the same env is also suitable for the worker
            # NOTE: shell escaping is a bit tricky here - careful on change!
            td.arguments = [raptor_file, raptor_class, self.uid]

            # ensure that defaults and backward compatibility kick in
            td.verify()

            # the default worker needs its own task description to derive the
            # amount of available resources
            self._reg['raptor.%s.cfg' % td.uid] = td.as_dict()

            # all workers run in the same sandbox as the master
            task = dict()

            task['origin']            = 'raptor'
            task['description']       = td.as_dict()
            task['state']             = rps.AGENT_STAGING_INPUT_PENDING
            task['status']            = self.NEW
            task['type']              = 'task'
            task['uid']               = td.uid
            task['task_sandbox_path'] = td.sandbox
            task['task_sandbox']      = 'file://localhost/' + td.sandbox
            task['pilot_sandbox']     = self._psbox
            task['session_sandbox']   = self._ssbox
            task['resource_sandbox']  = self._rsbox
            task['pilot']             = self._pid
            task['resources']         = {'cpu': td.ranks * td.cores_per_rank,
                                         'gpu': td.ranks * td.gpus_per_rank}
            tasks.append(task)

            # NOTE: the order of insert / state update relies on that order
            #       being maintained through the component's message push,
            #       the update worker's message receive up to the insertion
            #       order into the update worker's DB bulk op.
            self._log.debug('insert %s', td.uid)
            self.publish(rpc.STATE_PUBSUB, {'cmd': 'insert', 'arg': task})

            now = time.time()
            self._workers[td.uid] = {
                    'uid'        : td.uid,
                    'status'     : self.NEW,
                    'heartbeats' : {r: now for r in range(td.ranks)}
            }

        self.advance(tasks, publish=True, push=True)

        # dump registry with all worker descriptions ("raptor.<worker_uid>.cfg")
        self._reg.dump(self._uid)
        return [task['uid'] for task in tasks]


    # --------------------------------------------------------------------------
    #
    def wait_workers(self, count=None, uids=None):
        '''
        Wait for `n` workers, *or* for workers with given UID, *or* for all
        workers to become available, then return.  A worker is considered
        `available` when it registered with this master and get's its status
        flag set to `ACTIVE`.

        Args:

            count (int): number of workers to wait for
            uids (List[str]): set of worker UIDs to wait for
        '''

        if not uids and not count:
            # wait for all known workers by default
            uids = list(self._workers.keys())


        while True:

            # workers can be submitted by the client - we thus re-check for new
            # IDs on every loop
            if uids:
                check_uids = [uid for uid in uids if uid in self._workers]
            else:
                check_uids = list(self._workers.keys())

            if check_uids:
                stats = defaultdict(list)
                for uid in check_uids:
                    stats[self._workers[uid]['status']].append(uid)

                # if we wait for specific uids, check if all are ACTIVE
                if uids:
                    ok = True
                    for uid in check_uids:
                        if uid not in stats[self.ACTIVE]:
                            ok = False
                            break

                    if ok:
                        self._log.debug('wait ok')
                        return

                elif count:
                    if count <= len(stats[self.ACTIVE]):
                        self._log.debug('wait ok')
                        return

            if self._term.is_set():
                raise RuntimeError('wait interrupted by master termination')

            time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def start(self):
        '''
        start the main work thread of this master
        '''

        self._thread = mt.Thread(target=self._run)
        self._thread.daemon = True
        self._thread.start()


    # --------------------------------------------------------------------------
    #
    def stop(self):
        '''
        stop the main work thread of this master
        '''

        self._log.debug('set term from stop: %s', ru.get_stacktrace())
        self._term.set()

        super().stop()

        self.terminate()


    # --------------------------------------------------------------------------
    #
    def alive(self):
        '''
        check if the main work thread of this master is running
        '''

        if not self._thread or self._term.is_set():
            return False
        return True


    # --------------------------------------------------------------------------
    #
    def join(self):
        '''
        wait until the main work thread of this master completes
        '''

        if self._thread:
            self._thread.join()


    # --------------------------------------------------------------------------
    #
    def _hb_thread(self):
        '''
        main work threda of this master
        '''

        # wait for the submitted requests to complete
        while not self._term.is_set():

            self._log.debug('still alive')

            # check worker heartbeats
            now  = time.time()
            lost = set()
            for uid in self._workers:
                for rank, hb in self._workers[uid]['heartbeats'].items():
                    if hb < now - self._hb_tout:
                        self._log.warn('lost rank %d on worker %s', rank, uid)
                        lost.add(uid)

            time.sleep(self._hb_freq)

            for uid in lost:
                self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'worker_terminate',
                                                  'arg': {'uid': uid}})
                self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'cancel_tasks',
                                                  'arg': {'uids': [uid]}})


    # --------------------------------------------------------------------------
    #
    def _run(self):
        '''
        main work threda of this master
        '''

        hb_thread = mt.Thread(target=self._hb_thread)
        hb_thread.daemon = True
        hb_thread.start()

        # wait for the submitted requests to complete
        while not self._term.is_set():
            time.sleep(1)

        self._log.debug('terminate run loop')


    # --------------------------------------------------------------------------
    #
    def _run_task(self, td):
        '''
        accept a single task request for execution, execute it and wait for it's
        completion before returning the call.

        Note: this call is running in a separate thread created by an ZMQ
              Server instance and will thus not block the master's progress.
        '''

        # we get a dict but want a proper `TaskDescription` instance
        td = TaskDescription(td)

        # Create a new task ID for the submitted task (we do not allow
        # user-specified IDs in this case as we can't check uniqueness with the
        # client tmgr).  Then submit that task and wait for the `result_cb` to
        # report completion of the task

        tid   = '%s.%s' % (self.uid, ru.generate_id('subtask'))
        event = mt.Event()

        td['uid'] = tid
        task = Task(self, td, origin='raptor').as_dict()

        self._task_service_data[tid] = [event, task]
        self.submit_tasks([task])

        # wait for the result cb to pick up the td again
        event.wait()

        # update td info and remove data
        if len(self._task_service_data[tid]) != 3:
            self._log.error('invalid task_service data: %s',
                            self._task_service_data[tid])

        task.update(self._task_service_data[tid][2])

        del self._task_service_data[tid]

        # the task is completed and we can return it to the caller
        return task


    # --------------------------------------------------------------------------
    #
    def submit_tasks(self, tasks) -> None:
        '''
        submit a list of tasks to the task queue
        We expect to get either `TaskDescription` instances which will then get
        converted into task dictionaries and pushed out, or we get task
        dictionaries which are used as is.  Either way, `self.request_cb` will
        be called for all tasks submitted here.

        Args:
            tasks (List[TaskDescription]): description of tasks to be submitted
        '''

        normalized = list()
        for task in ru.as_list(tasks):

            if isinstance(task, TaskDescription):
                # convert to task dict
                normalized.append(Task(self, task, origin='raptor').as_dict())

            else:
                normalized.append(task)

        self._submit_tasks(self.request_cb(normalized))


    # --------------------------------------------------------------------------
    #
    def _submit_tasks(self, tasks) -> None:
        '''
        This is the internal implementation of `self.submit_tasks` which
        performs the actual submission after the tasks passed through the
        `request_cb`.
        '''

        if not tasks:
            return

        raptor_tasks     = list()
        executable_tasks = list()

        for task in ru.as_list(tasks):

            assert 'description' in task

            self._log.debug('submit: %s', task['uid'])

            mode = task['description'].get('mode', TASK_EXECUTABLE)
            if mode == TASK_EXECUTABLE:
                executable_tasks.append(task)
            else:
                raptor_tasks.append(task)

            # tasks submitted in raptor will miss sandbox completion as
            # performed by the tmgr.scheduler, so we add it here
            dummy = {'pilot_sandbox': self._psbox}
            sbox  = self._session._get_task_sandbox(task, dummy)
            task['task_sandbox']      = str(sbox)
            task['task_sandbox_path'] = ru.Url(sbox).path

        self._submit_executable_tasks(executable_tasks)
        self._submit_raptor_tasks(raptor_tasks)


    # --------------------------------------------------------------------------
    #
    def _submit_raptor_tasks(self, tasks) -> None:

        if tasks:

            self.advance(tasks, state=rps.AGENT_SCHEDULING,
                                publish=True, push=False)

            self._req_put.put(tasks)


    # --------------------------------------------------------------------------
    #
    def _submit_executable_tasks(self, tasks) -> None:
        '''
        Submit tasks per given task description to the agent this
        master is running in.
        '''

        if not tasks:
            return

        for task in tasks:

            td = task['description']

          # task['uid']               = td.get('uid')
            task['state']             = rps.AGENT_STAGING_INPUT_PENDING
            task['pilot_sandbox']     = self._psbox
            task['session_sandbox']   = self._ssbox
            task['resource_sandbox']  = self._rsbox
            task['pilot']             = self._pid
            task['resources']         = {'cpu': td['ranks'] *
                                                td.get('cores_per_rank', 1),
                                         'gpu': td['ranks'] *
                                                td.get('gpus_per_rank', 0.)}

            # NOTE: the order of insert / state update relies on that order
            #       being maintained through the component's message push,
            #       the update worker's message receive up to the insertion
            #       order into the update worker's DB bulk op.
            self._log.debug('insert %s', td['uid'])
            self.publish(rpc.STATE_PUBSUB, {'cmd': 'insert', 'arg': task})

        self.advance(tasks, state=rps.AGENT_STAGING_INPUT_PENDING,
                            publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def _request_cb(self, tasks) -> None:
        '''
        This cb will be called for all tasks which are under control of this
        raptor master, upon receival by the master.
        '''

        tasks = ru.as_list(tasks)

        for task in tasks:

            self._log.debug('request_cb: %s', task['uid'])

            # executable tasks which come from the scheduler are in danger of
            # entering a loop as we will push them back to the scheduler for
            # execution.  To avoid that loop we set an additional attribute
            # (raptor_seen=True) which the scheduler can filter for
            if task['description']['mode'] == TASK_EXECUTABLE:
                task['raptor_seen'] = True

        try:
            self.submit_tasks(tasks)

        except:
            self._log.exception('request cb failed')
            self.advance(tasks, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def request_cb(self, tasks):
        '''
        A raptor master implementation can overload this cb to filter all newly
        submitted tasks: it recieves a list of tasks and returns a potentially
        different list of tasks which are then executed.  It is up to the master
        implementation to ensure proper state transition for any tasks which
        are passed as argument but are not returned by the call and thus are not
        submitted for execution.

        Args:
            tasks ([List[Dict[str, ANY]]): list of tasks which this master
                received for execution

        Returns:
            tasks ([List[Dict[str, ANY]]): possibly different list of tasks than
                received

        '''

        # FIXME: document task format
        return tasks


    # --------------------------------------------------------------------------
    #
    def _result_cb(self, tasks):
        '''
        As pendant to the `_request_cb`, the `_result_cb` is invoked when raptor
        tasks complete execution.
        '''

        tasks = ru.as_list(tasks)

        for task in tasks:

            uid = task['uid']

            if not task.get('target_state'):

                ret = task.get('exit_code')
                if ret is None:
                    ret = -1

                if int(ret) == 0: task['target_state'] = rps.DONE
                else            : task['target_state'] = rps.FAILED

            if uid in self._task_service_data:

                # update task info and signal task service thread
                self._log.debug('unlock 2 %s', uid)
                self._task_service_data[uid].append(task)
                self._task_service_data[uid][0].set()

        try:
            self.result_cb(tasks)

        except:
            self._log.exception('result callback failed')

        self.advance(tasks, rps.AGENT_STAGING_OUTPUT_PENDING,
                            publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def result_cb(self, tasks):
        '''
        A raptor master implementation can overload this cb which get's called
        when raptor tasks complete execution.

        Args:
            tasks ([List[Dict[str, ANY]]): list of tasks which this master
                executed

        '''

        # FIXME: document task format
        pass


    # --------------------------------------------------------------------------
    #
    def terminate(self):
        '''
        terminate all workers and the master's own work loop.
        '''

        # unregister input queue
        self.publish(rpc.CONTROL_PUBSUB, {
                    'cmd': 'unregister_raptor_queue',
                    'arg': {'name' : self._uid,
                            'queue': self._input_queue.channel}})

        self._log.debug('set term from terminate')
        self._term.set()
        for uid in self._workers:
            self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'worker_terminate',
                                              'arg': {'uid': uid}})

        # wait for workers to terminate
        # FIXME: TS
      # self.wait()

        self._log.debug('all workers terminated')


# ------------------------------------------------------------------------------

